r"""
Matrix group over a ring that GAP understands
"""

# ****************************************************************************
#       Copyright (C) 2006      William Stein <wstein@gmail.com>
#                     2006-2008 David Joyner
#                     2013      Volker Braun <vbraun.name@gmail.com>
#                     2017      Dima Pasechnik
#                     2021      Frédéric Chapoton
#                     2018-2019 Sebastian Oehms
#                     2020      Vincent Delecroix
#                     2023      Matthias Koeppe
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.categories.groups import Groups
from sage.groups.group import Group
from sage.groups.libgap_mixin import GroupMixinLibGAP
from sage.groups.libgap_wrapper import ParentLibGAP
from sage.groups.matrix_gps.group_element import MatrixGroupElement_gap
from sage.groups.matrix_gps.matrix_group import MatrixGroup_generic


class MatrixGroup_gap(GroupMixinLibGAP, MatrixGroup_generic, ParentLibGAP):

    Element = MatrixGroupElement_gap

    def __init__(self, degree, base_ring, libgap_group, ambient=None, category=None):
        """
        Base class for matrix groups that implements GAP interface.

        INPUT:

        - ``degree`` -- integer; the degree (matrix size) of the
          matrix group

        - ``base_ring`` -- ring; the base ring of the matrices

        - ``libgap_group`` -- the defining libgap group

        - ``ambient`` -- a derived class of :class:`ParentLibGAP` or
          ``None`` (default); the ambient class if ``libgap_group``
          has been defined as a subgroup

        TESTS:

        ::

            sage: from sage.groups.matrix_gps.matrix_group_gap import MatrixGroup_gap
            sage: MatrixGroup_gap(2, ZZ, libgap.eval('GL(2, Integers)'))
            Matrix group over Integer Ring with 3 generators (
            [0 1]  [-1  0]  [1 1]
            [1 0], [ 0  1], [0 1]
            )

        Check that the slowness of GAP iterators and enumerators for matrix groups
        (cf. http://tracker.gap-system.org/issues/369) has been fixed::

            sage: i = iter(GL(6,5))
            sage: [ next(i) for j in range(8) ]
            [
            [1 0 0 0 0 0]  [4 0 0 0 0 1]  [0 4 0 0 0 0]  [0 4 0 0 0 0]
            [0 1 0 0 0 0]  [4 0 0 0 0 0]  [0 0 4 0 0 0]  [0 0 4 0 0 0]
            [0 0 1 0 0 0]  [0 4 0 0 0 0]  [0 0 0 4 0 0]  [0 0 0 4 0 0]
            [0 0 0 1 0 0]  [0 0 4 0 0 0]  [0 0 0 0 4 0]  [0 0 0 0 4 0]
            [0 0 0 0 1 0]  [0 0 0 4 0 0]  [0 0 0 0 0 4]  [0 0 0 0 0 4]
            [0 0 0 0 0 1], [0 0 0 0 4 0], [1 4 0 0 0 0], [2 4 0 0 0 0],
            [3 0 0 0 0 1]  [4 0 0 1 3 3]  [0 0 0 2 0 0]  [1 0 0 0 4 4]
            [3 0 0 0 0 0]  [4 0 0 0 3 3]  [0 0 0 0 4 0]  [1 0 0 0 0 4]
            [0 4 0 0 0 0]  [3 0 0 0 0 1]  [2 2 0 0 0 2]  [1 0 0 0 0 0]
            [0 0 4 0 0 0]  [3 0 0 0 0 0]  [1 4 0 0 0 0]  [0 1 0 0 0 0]
            [0 0 0 4 0 0]  [0 4 0 0 0 0]  [0 2 4 0 0 0]  [0 0 1 0 0 0]
            [4 0 0 0 2 3], [2 0 3 4 4 4], [0 0 1 4 0 0], [0 0 0 1 0 0]
            ]

        And the same for listing the group elements, as well as few other issues::

            sage: F = GF(3)
            sage: gens = [matrix(F,2, [1,0, -1,1]), matrix(F, 2, [1,1,0,1])]
            sage: G = MatrixGroup(gens)
            sage: G.cardinality()
            24
            sage: v = G.list()
            sage: len(v)
            24
            sage: v[:5]
            (
            [1 0]  [2 0]  [0 1]  [0 2]  [1 2]
            [0 1], [0 2], [2 0], [1 0], [2 2]
            )
            sage: all(g in G for g in G.list())
            True

        An example over a ring (see :issue:`5241`)::

            sage: M1 = matrix(ZZ,2,[[-1,0],[0,1]])
            sage: M2 = matrix(ZZ,2,[[1,0],[0,-1]])
            sage: M3 = matrix(ZZ,2,[[-1,0],[0,-1]])
            sage: MG = MatrixGroup([M1, M2, M3])
            sage: MG.list()
            (
            [1 0]  [ 1  0]  [-1  0]  [-1  0]
            [0 1], [ 0 -1], [ 0  1], [ 0 -1]
            )
            sage: MG.list()[1]
            [ 1  0]
            [ 0 -1]
            sage: MG.list()[1].parent()
            Matrix group over Integer Ring with 3 generators (
            [-1  0]  [ 1  0]  [-1  0]
            [ 0  1], [ 0 -1], [ 0 -1]
            )

        An example over a field (see :issue:`10515`)::

            sage: gens = [matrix(QQ,2,[1,0,0,1])]
            sage: MatrixGroup(gens).list()
            (
            [1 0]
            [0 1]
            )

        Another example over a ring (see :issue:`9437`)::

            sage: len(SL(2, Zmod(4)).list())
            48

        An error is raised if the group is not finite::

            sage: GL(2,ZZ).list()
            Traceback (most recent call last):
            ...
            NotImplementedError: group must be finite
        """
        ParentLibGAP.__init__(self, libgap_group, ambient=ambient)
        MatrixGroup_generic.__init__(self, degree, base_ring, category=category)

    def __iter__(self):
        """
        Iterate over the elements of the group.

        This method overrides the matrix group enumerator in GAP which
        does not (and often just cannot) work for infinite groups.

        TESTS:

        infinite groups can be dealt with::

            sage: import itertools
            sage: W = WeylGroup(["A",3,1])                                              # needs sage.rings.number_field
            sage: list(itertools.islice(W, int(4)))                                     # needs sage.rings.number_field
            [
            [1 0 0 0]  [-1  1  0  1]  [ 1  0  0  0]  [ 1  0  0  0]
            [0 1 0 0]  [ 0  1  0  0]  [ 1 -1  1  0]  [ 0  1  0  0]
            [0 0 1 0]  [ 0  0  1  0]  [ 0  0  1  0]  [ 0  1 -1  1]
            [0 0 0 1], [ 0  0  0  1], [ 0  0  0  1], [ 0  0  0  1]
            ]

        and finite groups, too::

            sage: G = GL(6,5)
            sage: list(itertools.islice(G, int(4)))
            [
            [1 0 0 0 0 0]  [4 0 0 0 0 1]  [0 4 0 0 0 0]  [0 4 0 0 0 0]
            [0 1 0 0 0 0]  [4 0 0 0 0 0]  [0 0 4 0 0 0]  [0 0 4 0 0 0]
            [0 0 1 0 0 0]  [0 4 0 0 0 0]  [0 0 0 4 0 0]  [0 0 0 4 0 0]
            [0 0 0 1 0 0]  [0 0 4 0 0 0]  [0 0 0 0 4 0]  [0 0 0 0 4 0]
            [0 0 0 0 1 0]  [0 0 0 4 0 0]  [0 0 0 0 0 4]  [0 0 0 0 0 4]
            [0 0 0 0 0 1], [0 0 0 0 4 0], [1 4 0 0 0 0], [2 4 0 0 0 0]
            ]
        """
        if not self.is_finite():
            # use implementation from category framework
            for g in super(Group, self).__iter__():
                yield g
            return
        # Use the standard GAP iterator for finite groups
        for g in super().__iter__():
            yield g
        return

    def _check_matrix(self, x_sage, x_gap):
        """
        Check whether the matrix ``x`` defines a group element.

        This is used by the element constructor (if you pass
        ``check=True``, the default) that the defining matrix is valid
        for this parent. Derived classes must override this to verify
        that the matrix is, for example, orthogonal or symplectic.

        INPUT:

        - ``x_sage`` -- a Sage matrix in the correct matrix space (degree
          and base ring)

        - ``x_gap`` -- the corresponding LibGAP matrix

        OUTPUT: a :exc:`TypeError` must be raised if ``x`` is invalid

        EXAMPLES::

            sage: m1 = matrix(GF(11), [(0, -1), (1, 0)])
            sage: m2 = matrix(GF(11), [(0, -1), (1, -1)])
            sage: G = MatrixGroup([m1, m2])
            sage: G([1,2,0,1])
            [1 2]
            [0 1]
            sage: G([1,1,1,0])
            Traceback (most recent call last):
            ...
            TypeError: matrix is not in the finitely generated group
        """
        from sage.libs.gap.libgap import libgap
        libgap_contains = libgap.eval(r'\in')
        is_contained = libgap_contains(x_gap, self.gap())
        if not is_contained.sage():
            raise TypeError('matrix is not in the finitely generated group')

    def subgroup(self, generators, check=True):
        """
        Return the subgroup generated by the given generators.

        INPUT:

        - ``generators`` -- list/tuple/iterable of group elements of ``self``
        - ``check`` -- boolean (default: ``True``); whether to check that each
          matrix is invertible

        OUTPUT: the subgroup generated by ``generators`` as an instance of
        :class:`FinitelyGeneratedMatrixGroup_gap`

        EXAMPLES::

            sage: # needs sage.rings.number_field
            sage: UCF = UniversalCyclotomicField()
            sage: G  = GL(3, UCF)
            sage: e3 = UCF.gen(3); e5 = UCF.gen(5)
            sage: m = matrix(UCF, 3,3, [[e3, 1, 0], [0, e5, 7],[4, 3, 2]])
            sage: S = G.subgroup([m]); S
            Subgroup with 1 generators (
            [E(3)    1    0]
            [   0 E(5)    7]
            [   4    3    2]
            ) of General Linear Group of degree 3 over Universal Cyclotomic Field

            sage: # needs sage.rings.number_field
            sage: CF3 = CyclotomicField(3)
            sage: G  = GL(3, CF3)
            sage: e3 = CF3.gen()
            sage: m = matrix(CF3, 3,3, [[e3, 1, 0], [0, ~e3, 7],[4, 3, 2]])
            sage: S = G.subgroup([m]); S
            Subgroup with 1 generators (
            [     zeta3          1          0]
            [         0 -zeta3 - 1          7]
            [         4          3          2]
            ) of General Linear Group of degree 3 over Cyclotomic Field of order 3 and degree 2

        TESTS::

            sage: TestSuite(G).run()                                                    # needs sage.rings.number_field
            sage: TestSuite(S).run()                                                    # needs sage.rings.number_field

            sage: # needs sage.rings.number_field
            sage: W = CoxeterGroup(['I',7])
            sage: s = W.simple_reflections()
            sage: G = W.subgroup([s[1]])
            sage: G.category()
            Category of finite groups

            sage: # needs sage.rings.number_field
            sage: W = WeylGroup(['A',2])
            sage: s = W.simple_reflections()
            sage: G = W.subgroup([s[1]])
            sage: G.category()
            Category of finite groups
        """
        # Override MatrixGroup_generic.subgroup
        return ParentLibGAP.subgroup(self, generators)

    def _subgroup_constructor(self, libgap_subgroup):
        """
        Return a finitely generated subgroup.

        See
        :meth:`sage.groups.libgap_wrapper.ParentLibGAP._subgroup_constructor`
        for details.

        TESTS::

            sage: SL2Z = SL(2,ZZ)
            sage: S, T = SL2Z.gens()
            sage: G = SL2Z.subgroup([T^2]); G   # indirect doctest
            Subgroup with 1 generators (
            [1 2]
            [0 1]
            ) of Special Linear Group of degree 2 over Integer Ring
            sage: G.ambient() is SL2Z
            True
        """
        cat = Groups()
        if self in Groups().Finite():
            cat = cat.Finite()
        from sage.groups.matrix_gps.finitely_generated_gap import FinitelyGeneratedMatrixGroup_gap
        return FinitelyGeneratedMatrixGroup_gap(self.degree(), self.base_ring(),
                                                libgap_subgroup, ambient=self,
                                                category=cat)

    from sage.groups.generic import structure_description



# distutils: libraries = NTL_LIBRARIES gmp m
# distutils: extra_compile_args = NTL_CFLAGS
# distutils: include_dirs = NTL_INCDIR
# distutils: library_dirs = NTL_LIBDIR
# distutils: extra_link_args = NTL_LIBEXTRA
# distutils: language = c++

r"""
Matrices over the `\GF{2}` via NTL

This class is only provided to have a complete NTL interface and for
comparison purposes. Sage's native matrices over `F_2` are much faster
for many problems like matrix multiplication and Gaussian elimination.

AUTHORS:

- Martin Albrecht <malb@informatik.uni-bremen.de> 2008-09: initial version
"""

# ****************************************************************************
#         Copyright (C) 2005 William Stein <wstein@gmail.com>
#    Copyright (C) 2008 Martin Albrecht <malb@informatik.uni-bremen.de>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from cysignals.signals cimport sig_on, sig_off
from sage.ext.cplusplus cimport ccrepr

include 'misc.pxi'
include 'decl.pxi'

from cpython.object cimport Py_EQ, Py_NE
from sage.libs.ntl.ntl_GF2 cimport ntl_GF2
from sage.rings.integer cimport Integer
from sage.libs.ntl.ntl_ZZ import unpickle_class_args


cdef class ntl_mat_GF2():
    r"""
    The \class{mat_GF2} class implements arithmetic with matrices over `F_2`.
    """
    def __init__(self, nrows=0, ncols=0, v=None):
        """
        Construct a matrix over ntl.GF2.

        INPUT:

        - ``nrows`` -- number of rows
        - ``ncols`` -- number of columns
        - ``v`` -- either a list or a matrix over GF(2^x)

        EXAMPLES::

            sage: A = ntl.mat_GF2(4,4); A
            [[0 0 0 0]
            [0 0 0 0]
            [0 0 0 0]
            [0 0 0 0]
            ]

            sage: A = random_matrix(GF(2),4,4); A  # random
            [0 1 0 1]
            [0 1 1 1]
            [0 0 0 1]
            [0 1 1 0]

            sage: B = ntl.mat_GF2(A); B  # random
            [[0 1 0 1]
            [0 1 1 1]
            [0 0 0 1]
            [0 1 1 0]
            ]

            sage: B = ntl.mat_GF2(4, 4, A.list())
            sage: B == A
            True
        """
        cdef Py_ssize_t _nrows, _ncols
        cdef Py_ssize_t i, j
        cdef GF2_c _elem

        from sage.structure.element import Matrix

        if isinstance(nrows, Matrix):
            _nrows = nrows.nrows()
            _ncols = nrows.ncols()
            v = nrows
            self.x.SetDims(_nrows, _ncols)
            sig_on()
            for i from 0 <= i < _nrows:
                for j from 0 <= j < _ncols:
                    GF2_conv_long(_elem, int(v[i, j]) % 2)
                    mat_GF2_setitem(&self.x, i, j, &_elem)
            sig_off()
            return

        _nrows = nrows
        _ncols = ncols
        self.x.SetDims(_nrows, _ncols)

        if v is not None:
            sig_on()
            for i from 0 <= i < _nrows:
                for j from 0 <= j < _ncols:
                    elem = v[i*_ncols+j]
                    if not isinstance(elem, ntl_GF2):
                        elem = ntl_GF2(elem)
                    mat_GF2_setitem(&self.x, i, j, &(<ntl_GF2>elem).x)
            sig_off()

    cdef ntl_GF2 _new_element(self):
        cdef ntl_GF2 r
        r = ntl_GF2.__new__(ntl_GF2)
        return r

    cdef ntl_mat_GF2 _new(self):
        cdef ntl_mat_GF2 r
        r = ntl_mat_GF2.__new__(ntl_mat_GF2)
        r.x.SetDims(self.x.NumRows(),self.x.NumCols())
        return r

    def __reduce__(self):
        """
        EXAMPLES::

            sage: A = random_matrix(GF(2),4,4)
            sage: B = ntl.mat_GF2(A)
            sage: loads(dumps(B)) == B # indirect doctest
            True
        """
        return unpickle_class_args, (ntl_mat_GF2, (self.x.NumRows(), self.x.NumCols(), self.list()))

    def __repr__(self):
        """
        Return the string representation of this matrix.

        EXAMPLES::

            sage: A = random_matrix(GF(2),4,4)
            sage: B = ntl.mat_GF2(A)
            sage: B.__repr__()[1:-2] == A.__repr__()
            True
        """
        return ccrepr(self.x)

    def __mul__(ntl_mat_GF2 self, other):
        """
        EXAMPLES::

            sage: A = random_matrix(GF(2),4,4)
            sage: B = random_matrix(GF(2),4,4)
            sage: c = ntl.mat_GF2(A)*ntl.mat_GF2(B)
            sage: c._sage_() == A*B
            True
        """
        cdef ntl_mat_GF2 r = self._new()
        if not isinstance(other, ntl_mat_GF2):
            other = ntl_mat_GF2(other)
        sig_on()
        mat_GF2_mul(r.x, self.x, (<ntl_mat_GF2>other).x)
        sig_off()
        return r

    def __sub__(ntl_mat_GF2 self, other):
        """
        EXAMPLES::

            sage: A = random_matrix(GF(2),4,4)
            sage: B = random_matrix(GF(2),4,4)
            sage: c = ntl.mat_GF2(A) - ntl.mat_GF2(B)
            sage: c._sage_() == A - B
            True
        """
        cdef ntl_mat_GF2 r = self._new()
        if not isinstance(other, ntl_mat_GF2):
            other = ntl_mat_GF2(other,)
        sig_on()
        mat_GF2_sub(r.x, self.x, (<ntl_mat_GF2>other).x)
        sig_off()
        return r

    def __add__(ntl_mat_GF2 self, other):
        """
        EXAMPLES::

            sage: A = random_matrix(GF(2),4,4)
            sage: B = random_matrix(GF(2),4,4)
            sage: c = ntl.mat_GF2(A) + ntl.mat_GF2(B)
            sage: c._sage_() == A + B
            True
        """
        cdef ntl_mat_GF2 r = self._new()
        if not isinstance(other, ntl_mat_GF2):
            other = ntl_mat_GF2(other)
        sig_on()
        mat_GF2_add(r.x, self.x, (<ntl_mat_GF2>other).x)
        sig_off()
        return r

    def __neg__(ntl_mat_GF2 self):
        """
        EXAMPLES::

            sage: A = random_matrix(GF(2),4,4)
            sage: (-ntl.mat_GF2(A))._sage_() == -A
            True
        """
        cdef ntl_mat_GF2 r = self._new()
        sig_on()
        mat_GF2_negate(r.x, self.x)
        sig_off()
        return r

    def __pow__(ntl_mat_GF2 self, long e, ignored):
        """
        EXAMPLES::

            sage: A = random_matrix(GF(2),4,4)
            sage: Abar = ntl.mat_GF2(A)
            sage: (Abar^0)._sage_() == A^0
            True
            sage: (Abar^1)._sage_() == A^1
            True
            sage: (Abar^2)._sage_() == A^2
            True
            sage: (Abar^3)._sage_() == A^3
            True
        """
        cdef ntl_mat_GF2 r = self._new()
        sig_on()
        mat_GF2_power(r.x, self.x, e)
        sig_off()
        return r

    def __richcmp__(ntl_mat_GF2 self, other, int op):
        """
        Compare ``self`` to ``other``.

        EXAMPLES::

            sage: A = random_matrix(GF(2),4,4)
            sage: A1 = ntl.mat_GF2(A)
            sage: A2 = ntl.mat_GF2(A)
            sage: A1 == A2
            True
            sage: A1[0,0] += 1
            sage: A1 == A2
            False
            sage: A1 == "x"
            False
        """
        if op != Py_EQ and op != Py_NE:
            raise TypeError("matrices over GF(2) are not ordered")

        cdef ntl_mat_GF2 b
        try:
            b = <ntl_mat_GF2?>other
        except TypeError:
            return NotImplemented

        return (op == Py_EQ) == (self.x == b.x)

    def NumRows(self):
        """
        Return the number of rows of this matrix.

        EXAMPLES::

            sage: A = ntl.mat_GF2(10,10)
            sage: A.NumRows()
            10
        """
        return int(self.x.NumRows())

    def NumCols(self):
        """
        Return the number of columns of this matrix.

        EXAMPLES::

            sage: A = ntl.mat_GF2(10,10)
            sage: A.NumCols()
            10
        """
        return int(self.x.NumCols())

    def __setitem__(self, ij, x):
        """
        EXAMPLES::

            sage: A = ntl.mat_GF2(5,5)
            sage: A[0,0] = 1
            sage: A[0,2] = 1
            sage: A
            [[1 0 1 0 0]
            [0 0 0 0 0]
            [0 0 0 0 0]
            [0 0 0 0 0]
            [0 0 0 0 0]
            ]
        """
        cdef int i, j
        if not isinstance(x, ntl_GF2):
            x = ntl_GF2(x)

        if isinstance(ij, tuple) and len(ij) == 2:
            i, j = ij
        elif self.x.NumCols() == 1 and isinstance(ij, (Integer, int)):
            i = ij
            j = 0
        elif self.x.NumRows() == 1 and isinstance(ij, (Integer, int)):
            i = 0
            j = ij
        else:
            raise TypeError('ij is not a matrix index')

        if i < 0 or i >= self.x.NumRows() or j < 0 or j >= self.x.NumCols():
            raise IndexError("array index out of range")

        mat_GF2_setitem(&self.x, i, j, &(<ntl_GF2>x).x)

    def __getitem__(self, ij):
        """
        EXAMPLES::

            sage: A = ntl.mat_GF2(3,3,range(9))
            sage: A[0,0]
            0
            sage: A[1,2]
            1
        """
        cdef int i, j
        if isinstance(ij, tuple) and len(ij) == 2:
            i, j = ij
        elif self.x.NumCols() == 1 and isinstance(ij, (Integer, int)):
            i = ij
            j = 0
        elif self.x.NumRows() == 1 and isinstance(ij, (Integer, int)):
            i = 0
            j = ij
        else:
            raise TypeError('ij is not a matrix index')

        if i < 0 or i >= self.x.NumRows() or j < 0 or j >= self.x.NumCols():
            raise IndexError("array index out of range")

        cdef ntl_GF2 e = self._new_element()
        e.x = self.x.get( i+1, j+1 )
        return e

    def determinant(self):
        """
        Return the determinant.

        EXAMPLES::

            sage: A = ntl.mat_GF2(3,3,range(9))
            sage: A.determinant()
            0
            sage: A = ntl.mat_GF2(3,3,[1,0,0, 0,1,0, 0,0,1])
            sage: A.determinant()
            1
        """
        cdef ntl_GF2 r = self._new_element()
        sig_on()
        r.x = mat_GF2_determinant(self.x)
        sig_off()
        return r

    def gauss(self, ncols=-1):
        r"""
        Perform unitary row operations so as to bring this matrix
        into row echelon form (not reduced!).

        If the optional argument ``ncols`` is supplied, stops when
        first ``ncols`` columns are in echelon form.  The return value is
        the rank (or the rank of the first ``ncols`` columns).

        INPUT:

        - ``ncols`` -- number of columns to process (default: all)

        EXAMPLES::

            sage: A = random_matrix(GF(2), 10, 10)
            sage: Abar = ntl.mat_GF2(A)
            sage: A.rank() == Abar.gauss()
            True
            sage: Abar  # random
            [[1 1 1 1 0 1 0 1 1 0]
            [0 1 1 1 0 1 1 0 0 1]
            [0 0 1 1 1 1 0 0 0 0]
            [0 0 0 1 0 0 1 1 1 1]
            [0 0 0 0 1 1 0 1 0 0]
            [0 0 0 0 0 1 1 1 0 1]
            [0 0 0 0 0 0 0 1 0 1]
            [0 0 0 0 0 0 0 0 0 1]
            [0 0 0 0 0 0 0 0 0 0]
            [0 0 0 0 0 0 0 0 0 0]
            ]

        ``Abar`` is in row echelon form now::

            sage: first_nonzero_indices = [Abar._sage_().row(i).nonzero_positions()[0] for i in range(A.rank())]
            sage: all(first_nonzero_indices[i] < first_nonzero_indices[i+1] for i in range(A.rank()-1))
            True

        ``Abar`` is not reduced::

            sage: all(Abar._sage_().row(i).nonzero_positions() == [] for i in range(A.rank(), Abar.NumRows()))
            True
        """
        if ncols == -1:
            ncols = self.x.NumCols()
        return int(mat_GF2_gauss(self.x, int(ncols)))

    def list(self):
        """
        Return a list of the entries in this matrix.

        EXAMPLES::

            sage: A = random_matrix(GF(2), 4, 4)
            sage: Abar = ntl.mat_GF2(A)
            sage: A.list() == Abar.list()
            True
        """
        cdef Py_ssize_t i, j
        return [self[i,j] for i in range(self.NumRows()) for j in range(self.x.NumCols())]

    def IsZero(self):
        r"""
        Return \code{True} if this matrix contains only zeroes, and \code{False} otherwise.

        EXAMPLES::

            sage: A = random_matrix(GF(2), 10, 10)
            sage: Abar = ntl.mat_GF2(A)
            sage: Abar.IsZero()
            False
            sage: Abar = ntl.mat_GF2(10,10)
            sage: Abar.IsZero()
            True
        """
        cdef long isZero
        sig_on()
        isZero = mat_GF2_IsZero(self.x)
        sig_off()
        return bool(isZero)

    def _sage_(ntl_mat_GF2 self):
        r"""
        Return a \class{Matrix} over GF(2).

        EXAMPLES::

            sage: A = random_matrix(GF(2), 6, 6)
            sage: Abar = ntl.mat_GF2(A)
            sage: Abar._sage_() == A
            True
        """
        from sage.rings.finite_rings.finite_field_constructor import FiniteField
        from sage.matrix.constructor import matrix
        m = matrix(FiniteField(2),self.x.NumRows(),self.x.NumCols())

        cdef Py_ssize_t i, j

        for i from 0 <= i < self.x.NumRows():
            for j from 0 <= j < self.x.NumCols():
                m[i,j] = GF2_conv_to_long(self.x.get( i+1, j+1))
        return m

    def transpose(ntl_mat_GF2 self):
        """
        Return the transposed matrix of this matrix.

        EXAMPLES::

            sage: A = random_matrix(GF(2), 10, 10)
            sage: Abar = ntl.mat_GF2(A)
            sage: Abar_t = Abar.transpose()
            sage: A_t = A.transpose()
            sage: A_t == Abar_t._sage_()
            True
        """
        cdef ntl_mat_GF2 r = self._new()
        sig_on()
        mat_GF2_transpose(r.x, self.x)
        sig_off()
        return r

    def __invert__(self):
        """
        Return `X = A^{-1}`; an error is raised if A is singular.

        EXAMPLES::

            sage: l = [0, 1, 1, 1, 1, 1, 0, 1, 0, 1, 1, 0, 1, 1, 1, 0, \
            ....:      0, 1, 1, 1, 1, 0, 0, 1, 1, 1, 1, 0, 1, 0, 1, 0, \
            ....:      1, 0, 1, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 1, 0, 0, \
            ....:      0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 1, 0, 1, 1, 0]
            sage: A = ntl.mat_GF2(8,8,l)
            sage: E = ~A*A
            sage: E.IsIdent()
            True
        """
        cdef ntl_mat_GF2 r = self._new()
        sig_on()
        mat_GF2_inv(r.x, self.x)
        sig_off()
        return r

    def IsIdent(self, n=-1):
        """
        Test if this matrix is the n x n identity matrix.

        EXAMPLES::

            sage: A = ntl.mat_GF2(4,4)
            sage: A[0,0] = 1
            sage: A[1,1] = 1
            sage: A[2,2] = 1
            sage: A.IsIdent()
            False
            sage: A[3,3] = 1
            sage: A.IsIdent()
            True
        """
        if n < 0:
            n = self.NumRows()
        return bool(mat_GF2_IsIdent(self.x, n))

    def IsDiag(self, long n, ntl_GF2 d):
        """
        Test if X is an  n x n diagonal matrix with d on diagonal.

        EXAMPLES::

            sage: A = ntl.mat_GF2(4,4)
            sage: A[0,0] = 1
            sage: A[1,1] = 1
            sage: A[2,2] = 1
            sage: A.IsDiag(3, ntl.GF2(1))
            False
            sage: A[3,3] = 1
            sage: A.IsDiag(4, ntl.GF2(1))
            True
        """
        return bool(mat_GF2_IsDiag(self.x, n, d.x))

    def image(self):
        """
        If A is this matrix and X the matrix returned by this function
        then, the rows of X are computed as basis of A's row space.
        X is in row echelon form.

        EXAMPLES::

            sage: A = random_matrix(GF(2),10,10)
            sage: Abar = ntl.mat_GF2(A)
            sage: A_image = A.image().matrix()
            sage: Abar_image =  Abar.image()._sage_()
            sage: A_image.row_space() == Abar_image.row_space()
            True

        X is in row echelon form::

            sage: first_nonzero_indices = [row.nonzero_positions()[0] for row in Abar_image.rows()]
            sage: all(first_nonzero_indices[i] < first_nonzero_indices[i+1] for i in range(Abar_image.nrows() - 1))
            True
        """
        cdef ntl_mat_GF2 X = self._new()
        sig_on()
        mat_GF2_image(X.x, self.x)
        sig_off()
        return X

    def kernel(self):
        """
        Compute a basis for the kernel of the map x -> x*A. where x
        is a row vector.

        EXAMPLES::

            sage: A = random_matrix(GF(2),10,10)
            sage: Abar = ntl.mat_GF2(A)
            sage: K_abar = Abar.kernel(); K_abar  # random
            [[0 0 0 1 1 0 1 0 1 0]
            [1 1 1 0 1 1 0 1 0 0]
            ]
            sage: (K_abar*Abar).IsZero()
            True
            sage: K_a = A.kernel().matrix()
            sage: K_a.row_space() == K_abar._sage_().row_space()
            True
        """
        cdef ntl_mat_GF2 X = self._new()
        sig_on()
        mat_GF2_kernel(X.x, self.x)
        sig_off()
        return X



r"""
Combinatorial species

A combinatorial species, as introduced by Joyal, is a functor from
the category of finite sets with bijections to the category of finite
sets with bijections.  Alternatively, we can regard a combinatorial
species as a formal sum of group actions of the symmetric groups
`\mathfrak S_n`, for `n\in\NN`.  For example, the trivial action of
`\mathfrak S_n` corresponds to the species `\mathcal E_n` of sets of
cardinality `n`.

More generally, a weighted multisort species in `k` sorts is a
functor from the category of `k`-tuples of finite sets with
bijections to the category of weighted finite sets with
weight-preserving bijections.  We may think of the sorts as
variables, traditionally denoted by `X, Y, Z`.

Such a species is equivalent to a formal sum of group actions of
groups `\mathfrak S_{n_1} \times \cdots \times \mathfrak S_{n_k}` with
`(n_1,\ldots,n_k)\in\NN^k`, together with a weight on the orbits of
each group action.  Yet more generally, a virtual weighted multisort
species is a formal difference of weighted multisort species.

We regard a combinatorial species as a sequence of group actions of
the symmetric groups `\mathfrak S_n`, for `n\in\NN`.

Coefficients of lazy species are computed on demand.  They have
infinite precision, although equality can only be decided in special
cases.

There are currently two implementations of combinatorial species, one
of which is deprecated and will be removed once all of the
functionality has been ported to the new framework.

The recommended implementation is :ref:`sage.rings.lazy_species` and
used as the following examples illustrate.

EXAMPLES:

We define rooted ordered trees with leaves having weight `q`::

    sage: R.<q> = QQ[]
    sage: L.<X> = LazyCombinatorialSpecies(R)
    sage: leaf = X
    sage: node = q * X
    sage: linorder = X/(1 - X)
    sage: T = L.undefined(valuation=1)
    sage: T.define(leaf + node * linorder(T))
    sage: T.isotype_generating_series()
    X + q*X^2 + ((q^2+q)*X^3) + ((q^3+3*q^2+q)*X^4) + ((q^4+6*q^3+6*q^2+q)*X^5)
    + ((q^5+10*q^4+20*q^3+10*q^2+q)*X^6) + O(X^7)

We define rooted unordered trees with leaves of sort `Y`.  The
standard representation of a species is its molecular decomposition::

    sage: L = LazyCombinatorialSpecies(R, "X")
    sage: E = L.Sets()
    sage: Ep = E.restrict(1)
    sage: M.<X, Y> = LazyCombinatorialSpecies(R)
    sage: A = M.undefined(valuation=1)
    sage: A.define(Y + X * Ep(A))
    sage: A.truncate(5)
    Y + X*Y + (X^2*Y+X*E_2(Y)) + (X^3*Y+X^2*E_2(Y)+X^2*Y^2+X*E_3(Y))

Introductory material
---------------------

- :ref:`section-examples-catalan`
- :ref:`section-generic-species`

Basic Species
-------------

- :ref:`sage.combinat.species.species`
- :ref:`sage.combinat.species.empty_species`
- :ref:`sage.combinat.species.recursive_species`
- :ref:`sage.combinat.species.characteristic_species`
- :ref:`sage.combinat.species.cycle_species`
- :ref:`sage.combinat.species.partition_species`
- :ref:`sage.combinat.species.permutation_species`
- :ref:`sage.combinat.species.linear_order_species`
- :ref:`sage.combinat.species.set_species`
- :ref:`sage.combinat.species.subset_species`
- :ref:`sage.combinat.species.library`

Operations on Species
---------------------

- :ref:`sage.combinat.species.sum_species`
- :ref:`sage.combinat.species.product_species`
- :ref:`sage.combinat.species.composition_species`
- :ref:`sage.combinat.species.functorial_composition_species`

Miscellaneous
-------------

- :ref:`sage.combinat.species.structure`
- :ref:`sage.combinat.species.misc`

"""
# install the docstring of this module to the containing package
from sage.misc.namespace_package import install_doc
install_doc(__package__, __doc__)

from sage.misc.lazy_import import lazy_import
lazy_import("sage.combinat.species.recursive_species", "CombinatorialSpecies",
            deprecation=(38544, "combinat.species is superseded by LazyCombinatorialSpecies"))

lazy_import("sage.combinat.species", "library", as_='species',
            deprecation=(38544, "combinat.species is superseded by LazyCombinatorialSpecies"))
del lazy_import
del install_doc



import sage.crypto.sbox
from sage.misc.persist import register_unpickle_override
register_unpickle_override('sage.crypto.mq.sbox', 'SBox', sage.crypto.sbox.SBox)

from sage.misc.lazy_import import lazy_import

lazy_import('sage.crypto.classical', ['AffineCryptosystem',
                                      'HillCryptosystem',
                                      'SubstitutionCryptosystem',
                                      'ShiftCryptosystem',
                                      'TranspositionCryptosystem',
                                      'VigenereCryptosystem',
                                      ])

lazy_import('sage.crypto.stream', ['LFSRCryptosystem',
                                   'ShrinkingGeneratorCryptosystem',
                                   ])

lazy_import('sage.crypto.lfsr', ['lfsr_sequence',
                                 'lfsr_autocorrelation',
                                 'lfsr_connection_polynomial',
                                 ])

lazy_import('sage.crypto.public_key.key_exchange', 'all', 'key_exchange')

del lazy_import



r"""
Carlitz module

AUTHORS:

- Xavier Caruso (2025-07): initial version
"""

# *****************************************************************************
#        Copyright (C) 2025 Xavier Caruso <xavier@caruso.ovh>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#                   http://www.gnu.org/licenses/
# *****************************************************************************

from sage.misc.cachefunc import cached_function

from sage.structure.parent import Parent
from sage.structure.element import Element
from sage.categories.finite_fields import FiniteFields

from sage.rings.infinity import Infinity

from sage.rings.polynomial.polynomial_ring import PolynomialRing_generic
from sage.rings.function_field.drinfeld_modules.drinfeld_module import DrinfeldModule


def CarlitzModule(A, base=None):
    r"""
    Return the Carlitz module over `A`.

    INPUT:

    - ``A`` -- a polynomial ring over a finite field

    - ``base`` -- a field, an element in a field or a
      string (default: the fraction field of ``A``)

    EXAMPLES::

        sage: Fq = GF(7)
        sage: A.<T> = Fq[]
        sage: CarlitzModule(A)
        Drinfeld module defined by T |--> τ + T

    We can specify a different base.
    This is interesting for instance for having two different variable
    names::

        sage: R.<z> = Fq[]
        sage: CarlitzModule(A, R)
        Drinfeld module defined by T |--> τ + z

    One can even use the following shortcut, which avoids the
    construction of `R`::

        sage: CarlitzModule(A, 'z')
        Drinfeld module defined by T |--> τ + z

    Using a similar syntax, we can construct the reduction of the
    Carlitz module modulo primes::

        sage: F.<a> = Fq.extension(z^2 + 1)
        sage: CarlitzModule(A, F)
        Drinfeld module defined by T |--> τ + a

    It is also possible to pass in any element in the base field
    (in this case, the result might not be strictly speaking the
    Carlitz module, but it is always a Drinfeld module of rank 1)::

        sage: CarlitzModule(A, z^2)
        Drinfeld module defined by T |--> τ + z^2

    TESTS::

        sage: CarlitzModule(Fq)
        Traceback (most recent call last):
        ...
        TypeError: the function ring must be defined over a finite field

    ::

        sage: S.<x,y> = QQ[]
        sage: CarlitzModule(A, S)
        Traceback (most recent call last):
        ...
        ValueError: function ring base must coerce into base field
    """
    if (not isinstance(A, PolynomialRing_generic)
     or A.base_ring() not in FiniteFields()):
        raise TypeError('the function ring must be defined over a finite field')
    if base is None:
        K = A.fraction_field()
        z = K.gen()
    elif isinstance(base, Parent):
        if base.has_coerce_map_from(A):
            z = base(A.gen())
        else:
            z = base.gen()
    elif isinstance(base, Element):
        z = base
    elif isinstance(base, str):
        K = A.base_ring()[base]
        z = K.gen()
    else:
        raise ValueError("cannot construct a Carlitz module from the given data")
    return DrinfeldModule(A, [z, 1])


def carlitz_exponential(A, prec=+Infinity, name='z'):
    r"""
    Return the Carlitz exponential attached the ring `A`.

    INPUT:

    - ``A`` -- a polynomial ring over a finite field

    - ``prec`` -- an integer or ``Infinity`` (default: ``Infinity``);
      the precision at which the series is returned; if ``Infinity``,
      a lazy power series in returned, else, a classical power series
      is returned.

    - ``name`` -- string (default: ``'z'``); the name of the
      generator of the lazy power series ring

    EXAMPLES::

        sage: A.<T> = GF(2)[]

     When ``prec`` is ``Infinity`` (which is the default),
     the exponential is returned as a lazy power series, meaning
     that any of its coefficients can be computed on demands::

        sage: exp = carlitz_exponential(A)
        sage: exp
        z + ((1/(T^2+T))*z^2) + ((1/(T^8+T^6+T^5+T^3))*z^4) + O(z^8)
        sage: exp[2^4]
        1/(T^64 + T^56 + T^52 + ... + T^27 + T^23 + T^15)
        sage: exp[2^5]
        1/(T^160 + T^144 + T^136 + ... + T^55 + T^47 + T^31)

    On the contrary, when ``prec`` is a finite number, all the
    required coefficients are computed at once::

        sage: carlitz_exponential(A, prec=10)
        z + (1/(T^2 + T))*z^2 + (1/(T^8 + T^6 + T^5 + T^3))*z^4 + (1/(T^24 + T^20 + T^18 + T^17 + T^14 + T^13 + T^11 + T^7))*z^8 + O(z^10)

    We check that the Carlitz exponential is the compositional inverse
    of the Carlitz logarithm::

        sage: log = carlitz_logarithm(A)
        sage: exp(log)
        z + O(z^8)
        sage: log(exp)
        z + O(z^8)
    """
    C = CarlitzModule(A)
    return C.exponential(prec, name)


def carlitz_logarithm(A, prec=+Infinity, name='z'):
    r"""
    Return the Carlitz exponential attached the ring `A`.

    INPUT:

    - ``A`` -- a polynomial ring over a finite field

    - ``prec`` -- an integer or ``Infinity`` (default: ``Infinity``);
      the precision at which the series is returned; if ``Infinity``,
      a lazy power series in returned, else, a classical power series
      is returned.

    - ``name`` -- string (default: ``'z'``); the name of the
      generator of the lazy power series ring

    EXAMPLES::

        sage: A.<T> = GF(2)[]

     When ``prec`` is ``Infinity`` (which is the default),
     the exponential is returned as a lazy power series, meaning
     that any of its coefficients can be computed on demands::

        sage: log = carlitz_logarithm(A)
        sage: log
        z + ((1/(T^2+T))*z^2) + ((1/(T^6+T^5+T^3+T^2))*z^4) + O(z^8)
        sage: log[2^4]
        1/(T^30 + T^29 + T^27 + ... + T^7 + T^5 + T^4)
        sage: log[2^5]
        1/(T^62 + T^61 + T^59 + ... + T^8 + T^6 + T^5)

    On the contrary, when ``prec`` is a finite number, all the
    required coefficients are computed at once::

        sage: carlitz_logarithm(A, prec=10)
        z + (1/(T^2 + T))*z^2 + (1/(T^6 + T^5 + T^3 + T^2))*z^4 + (1/(T^14 + T^13 + T^11 + T^10 + T^7 + T^6 + T^4 + T^3))*z^8 + O(z^10)
    """
    C = CarlitzModule(A)
    return C.logarithm(prec, name)


def carlitz_factorial(A, n):
    r"""
    Return the Carlitz factorial attached the ring `A`,
    evaluated at `n`.

    INPUT:

    - ``A`` -- a polynomial ring over a finite field

    - ``n`` -- an integer

    EXAMPLES::

        sage: A.<T> = GF(3)[]
        sage: carlitz_factorial(A, 0)
        1
        sage: carlitz_factorial(A, 1)
        T^3 + 2*T
        sage: carlitz_factorial(A, 2)
        T^6 + T^4 + T^2
        sage: carlitz_factorial(A, 3)
        T^18 + 2*T^12 + 2*T^10 + T^4

    TESTS::

        sage: carlitz_factorial(ZZ, 10)
        Traceback (most recent call last):
        ...
        TypeError: the function ring must be defined over a finite field
    """
    if (not isinstance(A, PolynomialRing_generic)
     or A.base_ring() not in FiniteFields()):
        raise TypeError('the function ring must be defined over a finite field')
    T = A.gen()
    q = A.base_ring().cardinality()
    ans = A.one()
    D = A.one()
    j = 1
    while n > 0:
        n, c = n.quo_rem(q)
        D = D**q * (T**(q**j) - T)
        ans *= D**c
        j += 1
    return ans


carlitz_series = {}


def carlitz_bernoulli(A, n):
    r"""
    Return the `n`-th Bernoulli-Carlitz number attached
    to the ring `A`.

    - ``A`` -- a polynomial ring over a finite field

    - ``n`` -- an integer

    EXAMPLES::

        sage: A.<T> = GF(3)[]
        sage: carlitz_bernoulli(A, 2)
        2*T^3 + T
        sage: carlitz_bernoulli(A, 4)
        T^15 + T^13 + T^11 + 2*T^7 + 2*T^5 + 2*T^3

    The Bernoulli-Carlitz numbers vanish when `n` is not a
    multiple of `q-1` where `q` is the cardinality of the
    base field::

        sage: carlitz_bernoulli(A, 1)
        0
        sage: carlitz_bernoulli(A, 3)
        0

    TESTS::

        sage: carlitz_bernoulli(ZZ, 10)
        Traceback (most recent call last):
        ...
        TypeError: the function ring must be defined over a finite field

    ::

        sage: B.<X> = GF(3)[]
        sage: carlitz_bernoulli(B, 2)
        2*X^3 + X
    """
    if (not isinstance(A, PolynomialRing_generic)
     or A.base_ring() not in FiniteFields()):
        raise TypeError('the function ring must be defined over a finite field')
    q = A.base_ring().cardinality()
    if q not in carlitz_series:
        e = carlitz_exponential(A)
        carlitz_series[q] = e.parent().gen() / e
    coeff = carlitz_series[q][n]
    coeff = A.fraction_field()(coeff)
    return A(coeff * carlitz_factorial(A, n))



# sage.doctest: needs sage.libs.flint sage.libs.pari
r"""
Creation of modular symbols spaces

EXAMPLES: We create a space and output its category.

::

    sage: C = HeckeModules(RationalField()); C
    Category of Hecke modules over Rational Field
    sage: M = ModularSymbols(11)
    sage: M.category()
    Category of Hecke modules over Rational Field
    sage: M in C
    True

We create a space compute the charpoly, then compute the same but
over a bigger field. In each case we also decompose the space using
`T_2`.

::

    sage: M = ModularSymbols(23,2, base_ring=QQ)
    sage: M.T(2).charpoly('x').factor()
    (x - 3) * (x^2 + x - 1)^2
    sage: M.decomposition(2)
    [Modular Symbols subspace of dimension 1 of Modular Symbols space of dimension 5 for Gamma_0(23) of weight 2 with sign 0 over Rational Field,
     Modular Symbols subspace of dimension 4 of Modular Symbols space of dimension 5 for Gamma_0(23) of weight 2 with sign 0 over Rational Field]

::

    sage: # needs sage.rings.number_field
    sage: M = ModularSymbols(23,2, base_ring=QuadraticField(5, 'sqrt5'))
    sage: M.T(2).charpoly('x').factor()
    (x - 3) * (x - 1/2*sqrt5 + 1/2)^2 * (x + 1/2*sqrt5 + 1/2)^2
    sage: M.decomposition(2)
    [Modular Symbols subspace of dimension 1 of Modular Symbols space of dimension 5 for Gamma_0(23) of weight 2 with sign 0 over Number Field in sqrt5 with defining polynomial x^2 - 5 with sqrt5 = 2.236067977499790?,
     Modular Symbols subspace of dimension 2 of Modular Symbols space of dimension 5 for Gamma_0(23) of weight 2 with sign 0 over Number Field in sqrt5 with defining polynomial x^2 - 5 with sqrt5 = 2.236067977499790?,
     Modular Symbols subspace of dimension 2 of Modular Symbols space of dimension 5 for Gamma_0(23) of weight 2 with sign 0 over Number Field in sqrt5 with defining polynomial x^2 - 5 with sqrt5 = 2.236067977499790?]

We compute some Hecke operators and do a consistency check::

    sage: m = ModularSymbols(39, 2)
    sage: t2 = m.T(2); t5 = m.T(5)
    sage: t2*t5 - t5*t2 == 0
    True

This tests the bug reported in :issue:`1220`::

    sage: G = GammaH(36, [13, 19])
    sage: G.modular_symbols()
    Modular Symbols space of dimension 13 for Congruence Subgroup Gamma_H(36)
     with H generated by [13, 19] of weight 2 with sign 0 over Rational Field
    sage: G.modular_symbols().cuspidal_subspace()
    Modular Symbols subspace of dimension 2 of Modular Symbols space of dimension 13 for
     Congruence Subgroup Gamma_H(36) with H generated by [13, 19] of weight 2 with sign 0
     over Rational Field

This test catches a tricky corner case for spaces with character::

    sage: ModularSymbols(DirichletGroup(20).1**3, weight=3, sign=1).cuspidal_subspace()
    Modular Symbols subspace of dimension 3 of Modular Symbols space of dimension 6
     and level 20, weight 3, character [1, -zeta4], sign 1,
     over Cyclotomic Field of order 4 and degree 2

This tests the bugs reported in :issue:`20932`::

    sage: chi = kronecker_character(3*34603)
    sage: ModularSymbols(chi, 2, sign=1, base_ring=GF(3))  # not tested  # long time (600 seconds)
    Modular Symbols space of dimension 11535 and level 103809, weight 2,
     character [2, 2], sign 1, over Finite Field of size 3
    sage: chi = kronecker_character(3*61379)
    sage: ModularSymbols(chi, 2, sign=1, base_ring=GF(3))  # not tested  # long time (1800 seconds)
    Modular Symbols space of dimension 20460 and level 184137, weight 2,
     character [2, 2], sign 1, over Finite Field of size 3
"""

# ****************************************************************************
#       Sage: Open Source Mathematical Software
#
#       Copyright (C) 2005 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

import weakref

from sage.categories.commutative_rings import CommutativeRings
from sage.categories.fields import Fields
import sage.modular.arithgroup.all as arithgroup
from sage.modular import dirichlet
from sage.rings.integer import Integer
from sage.rings.rational_field import RationalField


def canonical_parameters(group, weight, sign, base_ring):
    """
    Return the canonically normalized parameters associated to a choice
    of group, weight, sign, and base_ring. That is, normalize each of
    these to be of the correct type, perform all appropriate type
    checking, etc.

    EXAMPLES::

        sage: p1 = sage.modular.modsym.modsym.canonical_parameters(5,int(2),1,QQ) ; p1
        (Congruence Subgroup Gamma0(5), 2, 1, Rational Field)
        sage: p2 = sage.modular.modsym.modsym.canonical_parameters(Gamma0(5),2,1,QQ) ; p2
        (Congruence Subgroup Gamma0(5), 2, 1, Rational Field)
        sage: p1 == p2
        True
        sage: type(p1[1])
        <class 'sage.rings.integer.Integer'>
    """
    sign = Integer(sign)
    if sign not in [-1, 0, 1]:
        raise ValueError("sign must be -1, 0, or 1")

    weight = Integer(weight)
    if weight <= 1:
        raise ValueError("the weight must be at least 2")

    if isinstance(group, (int, Integer)):
        group = arithgroup.Gamma0(group)
    elif isinstance(group, dirichlet.DirichletCharacter):
        if group.is_trivial():
            group = arithgroup.Gamma0(group.modulus())
        else:
            eps = group.minimize_base_ring()
            group = (eps, eps.parent())
            if base_ring is None:
                base_ring = eps.base_ring()

    if base_ring is None:
        base_ring = RationalField()

    elif base_ring not in CommutativeRings():
        raise TypeError(f"base_ring (={base_ring}) must be a commutative ring")

    elif base_ring not in Fields():
        raise TypeError(f"(currently) base_ring (={base_ring}) must be a field")

    return group, weight, sign, base_ring


_cache = {}


def ModularSymbols_clear_cache():
    """
    Clear the global cache of modular symbols spaces.

    EXAMPLES::

        sage: sage.modular.modsym.modsym.ModularSymbols_clear_cache()
        sage: sorted(sage.modular.modsym.modsym._cache)
        []
        sage: M = ModularSymbols(6,2)
        sage: sorted(sage.modular.modsym.modsym._cache)
        [(Congruence Subgroup Gamma0(6), 2, 0, Rational Field)]
        sage: sage.modular.modsym.modsym.ModularSymbols_clear_cache()
        sage: sorted(sage.modular.modsym.modsym._cache)
        []

    TESTS:

    Make sure :issue:`10548` is fixed::

        sage: import gc
        sage: m = ModularSymbols(Gamma1(29))
        sage: m = []
        sage: ModularSymbols_clear_cache()
        sage: gc.collect() # random
        3422
        sage: a = [x for x in gc.get_objects()
        ....:      if isinstance(x,sage.modular.modsym.ambient.ModularSymbolsAmbient_wtk_g1)]
        sage: a
        []
    """
    global _cache
    _cache = {}


def ModularSymbols(group=1,
                   weight=2,
                   sign=0,
                   base_ring=None,
                   use_cache=True,
                   custom_init=None):
    r"""
    Create an ambient space of modular symbols.

    INPUT:

    - ``group`` -- a congruence subgroup or a Dirichlet character eps
    - ``weight`` -- integer; the weight, which must be >= 2
    - ``sign`` -- integer; the sign of the involution on modular symbols
      induced by complex conjugation. The default is 0, which means
      "no sign", i.e., take the whole space.
    - ``base_ring`` -- the base ring. Defaults to `\QQ` if no character
      is given, or to the minimal extension of `\QQ` containing the
      values of the character.
    - ``custom_init`` -- a function that is called with ``self`` as input
      before any computations are done using self; this could be used
      to set a custom modular symbols presentation.  If ``self`` is
      already in the cache and use_cache=True, then this function is
      not called.

    EXAMPLES: First we create some spaces with trivial character::

        sage: ModularSymbols(Gamma0(11),2).dimension()
        3
        sage: ModularSymbols(Gamma0(1),12).dimension()
        3

    If we give an integer N for the congruence subgroup, it defaults to
    `\Gamma_0(N)`::

        sage: ModularSymbols(1,12,-1).dimension()
        1
        sage: ModularSymbols(11,4, sign=1)
        Modular Symbols space of dimension 4 for Gamma_0(11) of weight 4
         with sign 1 over Rational Field

    We create some spaces for `\Gamma_1(N)`.

    ::

        sage: ModularSymbols(Gamma1(13),2)
        Modular Symbols space of dimension 15 for Gamma_1(13) of weight 2
         with sign 0 over Rational Field
        sage: ModularSymbols(Gamma1(13),2, sign=1).dimension()
        13
        sage: ModularSymbols(Gamma1(13),2, sign=-1).dimension()
        2
        sage: [ModularSymbols(Gamma1(7),k).dimension() for k in [2,3,4,5]]
        [5, 8, 12, 16]
        sage: ModularSymbols(Gamma1(5),11).dimension()
        20

    We create a space for `\Gamma_H(N)`::

        sage: G = GammaH(15,[4,13])
        sage: M = ModularSymbols(G,2)
        sage: M.decomposition()
        [Modular Symbols subspace of dimension 2 of Modular Symbols space of dimension 5 for Congruence Subgroup Gamma_H(15) with H generated by [4, 7] of weight 2 with sign 0 over Rational Field,
         Modular Symbols subspace of dimension 3 of Modular Symbols space of dimension 5 for Congruence Subgroup Gamma_H(15) with H generated by [4, 7] of weight 2 with sign 0 over Rational Field]

    We create a space with character::

        sage: e = (DirichletGroup(13).0)^2
        sage: e.order()
        6
        sage: M = ModularSymbols(e, 2); M
        Modular Symbols space of dimension 4 and level 13, weight 2, character [zeta6],
         sign 0, over Cyclotomic Field of order 6 and degree 2
        sage: f = M.T(2).charpoly('x'); f
        x^4 + (-zeta6 - 1)*x^3 - 8*zeta6*x^2 + (10*zeta6 - 5)*x + 21*zeta6 - 21
        sage: f.factor()
        (x - zeta6 - 2) * (x - 2*zeta6 - 1) * (x + zeta6 + 1)^2

    We create a space with character over a larger base ring than the values of the character::

        sage: # needs sage.rings.number_field
        sage: ModularSymbols(e, 2, base_ring=CyclotomicField(24))
        Modular Symbols space of dimension 4 and level 13, weight 2, character [zeta24^4],
         sign 0, over Cyclotomic Field of order 24 and degree 8

    More examples of spaces with character::

        sage: e = DirichletGroup(5, RationalField()).gen(); e
        Dirichlet character modulo 5 of conductor 5 mapping 2 |--> -1

        sage: m = ModularSymbols(e, 2); m
        Modular Symbols space of dimension 2 and level 5, weight 2, character [-1],
         sign 0, over Rational Field

    ::

        sage: m.T(2).charpoly('x')
        x^2 - 1
        sage: m = ModularSymbols(e, 6); m.dimension()
        6
        sage: m.T(2).charpoly('x')
        x^6 - 873*x^4 - 82632*x^2 - 1860496

    We create a space of modular symbols with nontrivial character in
    characteristic 2.

    ::

        sage: # needs sage.rings.finite_rings
        sage: G = DirichletGroup(13, GF(4,'a')); G
        Group of Dirichlet characters modulo 13
         with values in Finite Field in a of size 2^2
        sage: e = G.list()[2]; e
        Dirichlet character modulo 13 of conductor 13 mapping 2 |--> a + 1
        sage: M = ModularSymbols(e,4); M
        Modular Symbols space of dimension 8 and level 13, weight 4,
         character [a + 1], sign 0, over Finite Field in a of size 2^2
        sage: M.basis()
        ([X*Y,(1,0)], [X*Y,(1,5)], [X*Y,(1,10)], [X*Y,(1,11)],
         [X^2,(0,1)], [X^2,(1,10)], [X^2,(1,11)], [X^2,(1,12)])
        sage: M.T(2).matrix()
        [    0     0     0     0     0     0     1     1]
        [    0     0     0     0     0     0     0     0]
        [    0     0     0     0     0 a + 1     1     a]
        [    0     0     0     0     0     1 a + 1     a]
        [    0     0     0     0 a + 1     0     1     1]
        [    0     0     0     0     0     a     1     a]
        [    0     0     0     0     0     0 a + 1     a]
        [    0     0     0     0     0     0     1     0]

    We illustrate the custom_init function, which can be used to make
    arbitrary changes to the modular symbols object before its
    presentation is computed::

        sage: ModularSymbols_clear_cache()
        sage: def custom_init(M):
        ....:     M.customize='hi'
        sage: M = ModularSymbols(1,12, custom_init=custom_init)
        sage: M.customize
        'hi'

    We illustrate the relation between custom_init and use_cache::

        sage: def custom_init(M):
        ....:     M.customize='hi2'
        sage: M = ModularSymbols(1,12, custom_init=custom_init)
        sage: M.customize
        'hi'
        sage: M = ModularSymbols(1,12, custom_init=custom_init, use_cache=False)
        sage: M.customize
        'hi2'

    TESTS:

    We test use_cache::

        sage: ModularSymbols_clear_cache()
        sage: M = ModularSymbols(11,use_cache=False)
        sage: sage.modular.modsym.modsym._cache
        {}
        sage: M = ModularSymbols(11,use_cache=True)
        sage: sage.modular.modsym.modsym._cache
        {(Congruence Subgroup Gamma0(11), 2, 0, Rational Field): <weakref at ...; to '...ModularSymbolsAmbient_wt2_g0_with_category' at ...>}
        sage: M is ModularSymbols(11,use_cache=True)
        True
        sage: M is ModularSymbols(11,use_cache=False)
        False
    """
    from . import ambient
    key = canonical_parameters(group, weight, sign, base_ring)

    if use_cache and key in _cache:
        M = _cache[key]()
        if M is not None:
            return M

    group, weight, sign, base_ring = key

    M = None
    if isinstance(group, arithgroup.Gamma0_class):
        if weight == 2:
            M = ambient.ModularSymbolsAmbient_wt2_g0(
                group.level(), sign, base_ring, custom_init=custom_init)
        else:
            M = ambient.ModularSymbolsAmbient_wtk_g0(
                group.level(), weight, sign, base_ring, custom_init=custom_init)

    elif isinstance(group, arithgroup.Gamma1_class):

        M = ambient.ModularSymbolsAmbient_wtk_g1(group.level(),
                            weight, sign, base_ring, custom_init=custom_init)

    elif isinstance(group, arithgroup.GammaH_class):

        M = ambient.ModularSymbolsAmbient_wtk_gamma_h(group,
                            weight, sign, base_ring, custom_init=custom_init)

    elif isinstance(group, tuple):
        eps = group[0]
        M = ambient.ModularSymbolsAmbient_wtk_eps(eps,
                            weight, sign, base_ring, custom_init=custom_init)

    if M is None:
        raise NotImplementedError("computation of requested space of modular symbols not defined or implemented")

    if use_cache:
        _cache[key] = weakref.ref(M)
    return M



# distutils: libraries = GSL_LIBRARIES
# distutils: library_dirs = GSL_LIBDIR
# distutils: include_dirs = GSL_INCDIR
from sage.libs.gsl.types cimport *

cdef extern from "gsl/gsl_sf_legendre.h":

  double  gsl_sf_legendre_P1(double x)

  double  gsl_sf_legendre_P2(double x)

  double  gsl_sf_legendre_P3(double x)

  int  gsl_sf_legendre_P1_e(double x, gsl_sf_result * result)

  int  gsl_sf_legendre_P2_e(double x, gsl_sf_result * result)

  int  gsl_sf_legendre_P3_e(double x, gsl_sf_result * result)

  double  gsl_sf_legendre_Pl(int l, double x)

  int  gsl_sf_legendre_Pl_e(int l, double x, gsl_sf_result * result)

  int  gsl_sf_legendre_Pl_array(int lmax, double x, double result_array[])

  double  gsl_sf_legendre_Q0(double x)

  int  gsl_sf_legendre_Q0_e(double x, gsl_sf_result * result)

  double  gsl_sf_legendre_Q1(double x)

  int  gsl_sf_legendre_Q1_e(double x, gsl_sf_result * result)

  double  gsl_sf_legendre_Ql(int l, double x)

  int  gsl_sf_legendre_Ql_e(int l, double x, gsl_sf_result * result)

  double  gsl_sf_legendre_Plm(int l, int m, double x)

  int  gsl_sf_legendre_Plm_e(int l, int m, double x, gsl_sf_result * result)

  int  gsl_sf_legendre_Plm_array(int lmax, int m, double x, double result_array[])

  double  gsl_sf_legendre_sphPlm(int l, int m, double x)

  int  gsl_sf_legendre_sphPlm_e(int l, int m, double x, gsl_sf_result * result)

  int  gsl_sf_legendre_sphPlm_array(int lmax, int m, double x, double result_array[])

  int  gsl_sf_legendre_array_size(int lmax, int m)

  double  gsl_sf_conicalP_half(double lambd, double x)

  int  gsl_sf_conicalP_half_e(double lambd, double x, gsl_sf_result * result)

  double  gsl_sf_conicalP_mhalf(double lambd, double x)

  int  gsl_sf_conicalP_mhalf_e(double lambd, double x, gsl_sf_result * result)

  double  gsl_sf_conicalP_0(double lambd, double x)

  int  gsl_sf_conicalP_0_e(double lambd, double x, gsl_sf_result * result)

  double  gsl_sf_conicalP_1(double lambd, double x)

  int  gsl_sf_conicalP_1_e(double lambd, double x, gsl_sf_result * result)

  double  gsl_sf_conicalP_sph_reg(int l, double lambd, double x)

  int  gsl_sf_conicalP_sph_reg_e(int l, double lambd, double x, gsl_sf_result * result)

  double  gsl_sf_conicalP_cyl_reg(int m, double lambd, double x)

  int  gsl_sf_conicalP_cyl_reg_e(int m, double lambd, double x, gsl_sf_result * result)

  double  gsl_sf_legendre_H3d_0(double lambd, double eta)

  int  gsl_sf_legendre_H3d_0_e(double lambd, double eta, gsl_sf_result * result)

  double  gsl_sf_legendre_H3d_1(double lambd, double eta)

  int  gsl_sf_legendre_H3d_1_e(double lambd, double eta, gsl_sf_result * result)

  double  gsl_sf_legendre_H3d(int l, double lambd, double eta)

  int  gsl_sf_legendre_H3d_e(int l, double lambd, double eta, gsl_sf_result * result)

  int  gsl_sf_legendre_H3d_array(int lmax, double lambd, double eta, double result_array[])




r"""
Fractional morphisms of elliptic curves

Algorithms involving advanced computations with endomorphisms
or isogenies of elliptic curves sometimes involve divisions of
elliptic-curve morphisms by integers.
This operation yields another well-defined morphism whenever the
kernel of the numerator contains the kernel of the denominator.

The class :class:`EllipticCurveHom_fractional` represents symbolic
fractions `\varphi / n` where `\varphi\colon E\to E'` is any
:class:`~sage.schemes.elliptic_curves.hom.EllipticCurveHom` whose
kernel contains the `n`-torsion subgroup of `E`.
Functionality for converting this fraction to a more explicit form
is provided (:meth:`~EllipticCurveHom_fractional.to_isogeny_chain`).

EXAMPLES:

Division by an integer::

    sage: E = EllipticCurve(GF(419), [-1, 0])
    sage: phi = (E.frobenius_isogeny() + 1) / 2
    sage: phi
    Fractional elliptic-curve morphism of degree 105:
      Numerator:   Sum morphism:
        From: Elliptic Curve defined by y^2 = x^3 + 418*x over Finite Field of size 419
        To:   Elliptic Curve defined by y^2 = x^3 + 418*x over Finite Field of size 419
        Via:  (Frobenius endomorphism of degree 419:
            From: Elliptic Curve defined by y^2 = x^3 + 418*x over Finite Field of size 419
            To:   Elliptic Curve defined by y^2 = x^3 + 418*x over Finite Field of size 419,
          Scalar-multiplication endomorphism [1]
            of Elliptic Curve defined by y^2 = x^3 + 418*x over Finite Field of size 419)
      Denominator: 2
    sage: phi.degree()
    105
    sage: phi.to_isogeny_chain()
    Composite morphism of degree 105 = 1*3*5*7:
      From: Elliptic Curve defined by y^2 = x^3 + 418*x over Finite Field of size 419
      To:   Elliptic Curve defined by y^2 = x^3 + 418*x over Finite Field of size 419

Right division of isogenies::

    sage: E = EllipticCurve(GF(419), [1, 0])
    sage: ker = E(125, 70)
    sage: phi = E.isogeny(ker); phi
    Isogeny of degree 35
      from Elliptic Curve defined by y^2 = x^3 + x over Finite Field of size 419
      to Elliptic Curve defined by y^2 = x^3 + 289*x + 323 over Finite Field of size 419
    sage: psi = E.isogeny(5*ker); psi
    Isogeny of degree 7
      from Elliptic Curve defined by y^2 = x^3 + x over Finite Field of size 419
      to Elliptic Curve defined by y^2 = x^3 + 285*x + 87 over Finite Field of size 419
    sage: chi = phi.divide_right(psi); chi
    Fractional elliptic-curve morphism of degree 5:
      Numerator:   Composite morphism of degree 245 = 7*35:
      From: Elliptic Curve defined by y^2 = x^3 + 285*x + 87 over Finite Field of size 419
      To:   Elliptic Curve defined by y^2 = x^3 + 289*x + 323 over Finite Field of size 419
      Denominator: 7
    sage: phi == chi * psi
    True

Left division of isogenies::

    sage: E = EllipticCurve(GF(419), [1, 0])
    sage: ker = E(125, 70)
    sage: phi = E.isogeny(ker); phi
    Isogeny of degree 35
      from Elliptic Curve defined by y^2 = x^3 + x over Finite Field of size 419
      to Elliptic Curve defined by y^2 = x^3 + 289*x + 323 over Finite Field of size 419
    sage: tmp = E.isogeny(7*ker)
    sage: psi = tmp.codomain().isogeny(tmp(ker)); psi
    Isogeny of degree 7
      from Elliptic Curve defined by y^2 = x^3 + 269*x + 82 over Finite Field of size 419
      to Elliptic Curve defined by y^2 = x^3 + 289*x + 323 over Finite Field of size 419
    sage: chi = phi.divide_left(psi); chi
    Fractional elliptic-curve morphism of degree 5:
      Numerator:   Composite morphism of degree 245 = 35*7:
      From: Elliptic Curve defined by y^2 = x^3 + x over Finite Field of size 419
      To:   Elliptic Curve defined by y^2 = x^3 + 269*x + 82 over Finite Field of size 419
      Denominator: 7
    sage: phi == psi * chi
    True

AUTHORS:

- Lorenz Panny (2024)
"""
from sage.misc.cachefunc import cached_method
from sage.structure.sequence import Sequence

from sage.rings.integer_ring import ZZ

from sage.schemes.elliptic_curves.hom import EllipticCurveHom


class EllipticCurveHom_fractional(EllipticCurveHom):
    r"""
    This class represents a (symbolic) quotient of an isogeny divided by an integer.

    EXAMPLES::

        sage: from sage.schemes.elliptic_curves.hom_fractional import EllipticCurveHom_fractional
        sage: phi = EllipticCurve([1,1]).scalar_multiplication(-2)
        sage: EllipticCurveHom_fractional(phi, 2)
        Fractional elliptic-curve morphism of degree 1:
          Numerator:   Scalar-multiplication endomorphism [-2]
                         of Elliptic Curve defined by y^2 = x^3 + x + 1 over Rational Field
          Denominator: 2
        sage: EllipticCurveHom_fractional(phi, 3)
        Traceback (most recent call last):
        ...
        ValueError: Scalar-multiplication endomorphism [-2]
          of Elliptic Curve defined by y^2 = x^3 + x + 1 over Rational Field
          is not divisible by 3
    """

    def __init__(self, phi, d, *, check=True) -> None:
        r"""
        Construct a (symbolic) quotient of an isogeny divided by an integer.

        EXAMPLES::

            sage: from sage.schemes.elliptic_curves.hom_fractional import EllipticCurveHom_fractional
            sage: phi = EllipticCurve(GF(11), [1,1]).scalar_multiplication(-3)
            sage: EllipticCurveHom_fractional(phi, 3)
            Fractional elliptic-curve morphism of degree 1:
              Numerator:   Scalar-multiplication endomorphism [-3]
                             of Elliptic Curve defined by y^2 = x^3 + x + 1 over Finite Field of size 11
              Denominator: 3
        """
        if not isinstance(phi, EllipticCurveHom):
            raise TypeError('not an elliptic-curve morphism')

        d = ZZ(d)
        if d < 0:
            phi = -phi
            d = -d
        if not d:
            raise ZeroDivisionError('cannot divide isogeny by zero')

        if check:
            if not (d**2).divides(phi.degree()):
                raise ValueError(f'{phi} is not divisible by {d}')

            E = phi.domain()
            for l, e in d.factor():
                F = E.division_field(l**e)
                EE = E.change_ring(F)

                psi = E.division_polynomial(l**e).radical()
                psi //= psi.gcd(E.division_polynomial(l**(e-1)))
                if psi.degree() < l**2//2:
                    assert l == E.base_field().characteristic()
                    if psi.is_one():
                        # supersingular, hence [p] is the only p^2-isogeny up to isomorphism
                        continue
                    # ordinary, hence [p] is Frobenius times its dual
                    insep = phi.inseparable_degree().valuation(l)
                    sep = phi.degree().valuation(l) - insep
                    if sep < e or insep < e:
                        raise ValueError(f'{phi} is not divisible by {l**e}')
                    continue

                P, Q = _torsion_gens(E, EE, l, e, psi=psi)

                if phi._eval(P) or phi._eval(Q):
                    raise ValueError(f'{phi} is not divisible by {l**e}')

        self._phi = phi
        self._d = d
        self._degree = self._phi.degree() // d**2
        self._domain = phi.domain()
        self._codomain = phi.codomain()
        EllipticCurveHom.__init__(self, self._domain, self._codomain)

    def _call_(self, P):
        r"""
        Evaluate this fractional elliptic-curve morphism at a point.

        EXAMPLES::

            sage: E = EllipticCurve(GF(13), [9,0])
            sage: pi = E.frobenius_isogeny()
            sage: phi = (1 + pi) / 2
            sage: phi.degree()
            2
            sage: P, Q = E(6,7), E(2, 0)
            sage: phi(P)
            (4 : 3 : 1)
            sage: phi(Q)
            (0 : 0 : 1)
        """
        return self._eval(P)

    def _eval(self, P):
        r"""
        Less strict evaluation method for internal use.

        In particular, this can be used to evaluate ``self`` at a
        point defined over an extension field.

        INPUT: a sequence of 3 coordinates defining a point on ``self``

        OUTPUT: the result of evaluating ``self`` at the given point

        EXAMPLES::

            sage: E = EllipticCurve(GF(13), [9,0])
            sage: pi = E.frobenius_isogeny()
            sage: phi = (1 + pi) / 2
            sage: phi.degree()
            2
            sage: EE = E.change_ring(GF(13^3))
            sage: EE.base_field().inject_variables()
            Defining z3
            sage: P = EE(4*z3^2 + 11*z3, 8*z3^2 + 11*z3 + 8)
            sage: Q = EE(z3^2 + 7*z3 + 6, 3*z3^2 + 7*z3 + 10)
            sage: phi._eval(P)
            (9*z3^2 + 6*z3 + 6 : 4*z3^2 + 9*z3 + 3 : 1)
            sage: phi._eval(Q)
            (z3^2 + 9*z3 : 10*z3^2 + 6*z3 + 10 : 1)

        TESTS:

        Check for :issue:`41902`::

            sage: F.<t> = GF((1019, 2))
            sage: E = EllipticCurve(F, [1, 0])
            sage: P = E.lift_x(675*t + 800)
            sage: i = E.automorphisms()[-1]
            sage: j = E.frobenius_isogeny()
            sage: f = End(E)(6) / 6
            sage: P
            (675*t + 800 : 518*t + 493 : 1)
            sage: f(P)
            (675*t + 800 : 518*t + 493 : 1)

        Check for :issue:`41969`::

            sage: E, endo = special_supersingular_curve(GF(1021^2), endomorphism=True)
            sage: ((1 + endo + E.frobenius_isogeny()) / 2)((148, 0))
            (148 : 0 : 1)
        """
        if self._domain.defining_polynomial()(*P):
            raise ValueError(f'{P} not on {self._domain}')
        F = Sequence(P).universe()

        if not P:
            return self._codomain.base_extend(F).zero()

        P = P.divide(self._d, extend=True)
        Q = self._phi._eval(P).change_ring(F)

        return self._codomain.base_extend(F)(*Q)

    def _repr_(self) -> str:
        r"""
        Return a description of this fractional elliptic-curve morphism.

        EXAMPLES::

            sage: E = EllipticCurve(GF(13), [9,0])
            sage: pi = E.frobenius_isogeny()
            sage: (1 + pi) / 2
            Fractional elliptic-curve morphism of degree 2:
              Numerator:   Sum morphism:
                From: Elliptic Curve defined by y^2 = x^3 + 9*x over Finite Field of size 13
                To:   Elliptic Curve defined by y^2 = x^3 + 9*x over Finite Field of size 13
                Via:  (Scalar-multiplication endomorphism [1]
                         of Elliptic Curve defined by y^2 = x^3 + 9*x over Finite Field of size 13,
                       Frobenius endomorphism of degree 13:
                         From: Elliptic Curve defined by y^2 = x^3 + 9*x over Finite Field of size 13
                         To:   Elliptic Curve defined by y^2 = x^3 + 9*x over Finite Field of size 13)
              Denominator: 2
        """
        return f'Fractional elliptic-curve morphism of degree {self._degree}:' \
            f'\n  Numerator:   {self._phi}' \
            f'\n  Denominator: {self._d}'

    @cached_method
    def to_isogeny_chain(self):
        r"""
        Convert this fractional elliptic-curve morphism into a (non-fractional)
        :class:`~sage.schemes.elliptic_curves.hom_composite.EllipticCurveHom_composite`
        object representing the same morphism.

        EXAMPLES::

            sage: p = 419
            sage: E = EllipticCurve(GF(p^2), [1,0])
            sage: iota = E.automorphisms()[2]   # sqrt(-1)
            sage: pi = E.frobenius_isogeny()    # sqrt(-p)
            sage: endo = (iota + pi) / 2
            sage: endo.degree()
            105
            sage: endo.to_isogeny_chain()
            Composite morphism of degree 105 = 1*3*5*7:
              From: Elliptic Curve defined by y^2 = x^3 + x over Finite Field in z2 of size 419^2
              To:   Elliptic Curve defined by y^2 = x^3 + x over Finite Field in z2 of size 419^2
            sage: endo.to_isogeny_chain() == endo
            True
        """
        E = self._domain

        ker = []
        insep = 0
        for l, e in self._phi.degree().factor():
            F = E.division_field(l**e)
            EE = E.change_ring(F)

            psi = E.division_polynomial(l**e).radical()
            psi //= psi.gcd(E.division_polynomial(l**(e-1)))
            if psi.degree() < l**2//2:
                assert l == E.base_field().characteristic()
                if psi.is_one():
                    # supersingular, hence [p] is the only p^2-isogeny up to isomorphism
                    insep += 2*self._d.valuation(l)
                else:
                    # ordinary, hence [p] is Frobenius times its dual
                    insep += self._d.valuation(l)
                    ker.append(self._d.p_primary_part(l) * EE.lift_x(psi.any_root(ring=F)))
                continue

            P, Q = _torsion_gens(E, EE, l, e, psi=psi)
            if self.is_endomorphism():
                RS = None
            else:
                RS = _torsion_gens(self._codomain, self._codomain.change_ring(F), l, e)

            mat = self._phi.matrix_on_subgroup((P, Q), RS)
            for row in filter(bool, self._d.p_primary_part(l) * mat.left_kernel_matrix()):
                K = sum(ZZ(c)*T for c, T in zip(row, (P, Q)))
                K.set_order(multiple=l**e)
                assert self._eval(K) == 0
                ker.append(K)

        from sage.schemes.elliptic_curves.hom_composite import EllipticCurveHom_composite
        chain = EllipticCurveHom_composite(E, [])
        ker = ker[::-1]
        while ker:
            if not (P := ker.pop()):
                continue
            (l, e), = P.order().factor()
            K = l**(e-1)*P
            if e > 1:
                ker.append(P)
            poly = E.kernel_polynomial_from_point(K, algorithm='basic')  # FIXME algorithm='basic' is a workaround for #34907
            step = E.isogeny(poly)
            chain = step * chain
            ker = [step._eval(T) for T in ker]
            E = chain.codomain()

        for iso in E.isomorphisms(self._codomain):
            if self.scaling_factor() == iso.scaling_factor() * chain.scaling_factor():
                break
        else:
            assert False, 'bug in converting fractional isogeny to isogeny chain'

        return iso * chain

    # EllipticCurveHom methods

    @staticmethod
    def _composition_impl(left, right):
        r"""
        Specialized composition method for fractional elliptic-curve morphisms.

        TESTS::

            sage: E = EllipticCurve(GF(419), [-1,0])
            sage: pi = E.frobenius_isogeny()
            sage: endo = (1 + pi) / 2
            sage: 12 * endo
            Composite morphism of degree 15120 = 420*36:
              From: Elliptic Curve defined by y^2 = x^3 + 418*x over Finite Field of size 419
              To:   Elliptic Curve defined by y^2 = x^3 + 418*x over Finite Field of size 419
        """
        from sage.schemes.elliptic_curves.hom_scalar import EllipticCurveHom_scalar
        if isinstance(left, EllipticCurveHom_scalar):
            left, right = right, left
        if isinstance(left, EllipticCurveHom_fractional) and isinstance(right, EllipticCurveHom_scalar):
            r = right._m / left._d
            num, den = r.numerator(), r.denominator()
            if num.is_one():
                f = left._phi
            elif (-num).is_one():
                f = -left._phi
            else:
                f = num * left._phi
            if den.is_one():
                return f
            return EllipticCurveHom_fractional(f, den)
        return NotImplemented

    @staticmethod
    def _comparison_impl(left, right, op):
        r"""
        Compare a fractional elliptic-curve morphism to another elliptic-curve morphism.

        Called by :meth:`EllipticCurveHom._richcmp_`.

        TESTS::

            sage: E = EllipticCurve(GF(419), [-1,0])
            sage: pi = E.frobenius_isogeny()
            sage: endo = (1 + pi) / 2
            sage: 2 * endo - 1 == pi
            True
            sage: 1 - 2 * endo == pi
            False
        """
        assert isinstance(left, EllipticCurveHom_fractional) or isinstance(right, EllipticCurveHom_fractional)
        if isinstance(left, EllipticCurveHom_fractional):
            right = left._d * right
            left = left._phi
        if isinstance(right, EllipticCurveHom_fractional):
            left = right._d * left
            right = right._phi
        return EllipticCurveHom._richcmp_(left, right, op)

    def rational_maps(self):
        r"""
        Return the pair of explicit rational maps defining this fractional isogeny.

        EXAMPLES::

            sage: E = EllipticCurve(GF(419), [1,0])
            sage: phi = E.isogeny(E(185, 73)); phi.rational_maps()
            ((x^5 + 189*x^4 + 9*x^3 + 114*x^2 + 11*x + 206)/(x^4 + 189*x^3 - 105*x^2 - 171*x - 155),
             (x^6*y + 74*x^5*y - 127*x^4*y + 148*x^3*y + 182*x^2*y + 115*x*y + 43*y)/(x^6 + 74*x^5 - 13*x^4 + x^3 - 88*x^2 - 157*x - 179))
            sage: ((phi + phi) / 2).rational_maps()
            ((x^5 + 189*x^4 + 9*x^3 + 114*x^2 + 11*x + 206)/(x^4 + 189*x^3 - 105*x^2 - 171*x - 155),
             (x^6*y + 74*x^5*y - 127*x^4*y + 148*x^3*y + 182*x^2*y + 115*x*y + 43*y)/(x^6 + 74*x^5 - 13*x^4 + x^3 - 88*x^2 - 157*x - 179))
        """
        return self.to_isogeny_chain().rational_maps()

    def x_rational_map(self):
        r"""
        Return the `x`-coordinate rational map of this fractional isogeny.

        EXAMPLES::

            sage: E = EllipticCurve(GF(419), [1,0])
            sage: phi = E.isogeny(E(185, 73)); phi.x_rational_map()
            (x^5 + 189*x^4 + 9*x^3 + 114*x^2 + 11*x + 206)/(x^4 + 189*x^3 + 314*x^2 + 248*x + 264)
            sage: ((phi + phi) / 2).x_rational_map()
            (x^5 + 189*x^4 + 9*x^3 + 114*x^2 + 11*x + 206)/(x^4 + 189*x^3 + 314*x^2 + 248*x + 264)
        """
        return self.to_isogeny_chain().x_rational_map()

    def kernel_polynomial(self):
        r"""
        Return the kernel polynomial of this fractional isogeny.

        EXAMPLES::

            sage: E = EllipticCurve(GF(419), [1,0])
            sage: phi = E.isogeny(E(185, 73)); phi.kernel_polynomial()
            x^2 + 304*x + 39
            sage: ((phi + phi) / 2).kernel_polynomial()
            x^2 + 304*x + 39
        """
        return self.to_isogeny_chain().kernel_polynomial()

    @cached_method
    def dual(self):
        r"""
        Return the dual of this fractional isogeny.

        EXAMPLES::

            sage: E = EllipticCurve(GF(419), [1,0])
            sage: phi = E.isogeny(E(185, 73))
            sage: ((phi + phi) / 2).dual() == phi.dual()
            True
        """
        psi = EllipticCurveHom_fractional(self._phi.dual(), self._d)
        psi.dual.set_cache(self)
        return psi

    def formal(self, *args):
        r"""
        Return the formal isogeny corresponding to this fractional
        isogeny as a power series in the variable `t=-x/y` on the
        domain curve.

        EXAMPLES::

            sage: E = EllipticCurve(GF(419), [-1,0])
            sage: pi = E.frobenius_isogeny()
            sage: ((1 + pi) / 2).formal()
            210*t + 26*t^5 + 254*t^9 + 227*t^13 + 36*t^17 + 74*t^21 + O(t^23)
        """
        return self.to_isogeny_chain().formal(*args)

    def scaling_factor(self):
        r"""
        Return the Weierstrass scaling factor associated to this
        fractional isogeny.

        The scaling factor is the constant `u` (in the base field)
        such that `\varphi^* \omega_2 = u \omega_1`, where
        `\varphi: E_1\to E_2` is this morphism and `\omega_i` are
        the standard Weierstrass differentials on `E_i` defined by
        `\mathrm dx/(2y+a_1x+a_3)`.

        EXAMPLES::

            sage: E = EllipticCurve(GF(419), [-1,0])
            sage: pi = E.frobenius_isogeny()
            sage: ((1 + pi) / 2).scaling_factor()
            210
        """
        # FIXME this can crash when p | d
        return self._phi.scaling_factor() / self._d

    def inseparable_degree(self):
        r"""
        Return the inseparable degree of this morphism.

        EXAMPLES::

            sage: E = EllipticCurve(GF(419), [-1,0])
            sage: pi = E.frobenius_isogeny()
            sage: ((1 + pi) / 2).inseparable_degree()
            1
            sage: E = EllipticCurve(GF(419), [-1,0])
            sage: pi = E.frobenius_isogeny()
            sage: ((3*pi - pi) / 2).inseparable_degree()
            419
        """
        return self._phi.inseparable_degree() / self._domain.scalar_multiplication(self._d).inseparable_degree()


def _torsion_gens(E, EE, l, e, psi=None):
    if psi is None:
        psi = E.division_polynomial(l**e).radical()
        psi //= psi.gcd(E.division_polynomial(l**(e-1)))

    xs = iter(psi.roots(ring=EE.base_field(), multiplicities=False))
    P = EE.lift_x(next(xs))
    while True:
        Q = EE.lift_x(next(xs))
        if not (P.weil_pairing(Q, l**e)**(l**(e-1))).is_one():
            break
    else:
        assert False, f'bug in finding {l**e}-torsion basis'
    return P, Q



r"""
Points on projective schemes

This module implements scheme morphism for points on weighted projective schemes.

AUTHORS:

- Gareth Ma (2024): initial version, based on unweighted version.
"""

# ****************************************************************************
#       Copyright (C) 2006 David Kohel <kohel@maths.usyd.edu.au>
#       Copyright (C) 2006 William Stein <wstein@gmail.com>
#       Copyright (C) 2011 Volker Braun <vbraun.name@gmail.com>
#       Copyright (C) 2024 Gareth Ma <grhkm21@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.misc.misc_c import prod
from sage.structure.richcmp import op_EQ, op_NE, richcmp

from sage.categories.fields import Fields
from sage.categories.commutative_rings import CommutativeRings
from sage.misc.latex import latex
from sage.rings.fraction_field import FractionField
from sage.schemes.generic.morphism import SchemeMorphism, SchemeMorphism_point
from sage.structure.sequence import Sequence

# --------------------
# Projective varieties
# --------------------


class SchemeMorphism_point_weighted_projective_ring(SchemeMorphism_point):
    """
    A rational point of weighted projective space over a ring.

    INPUT:

    - ``X`` -- a homset of a subscheme of an ambient weighted projective space over a ring `R`.

    - ``v`` -- a list or tuple of coordinates in `R`.

    - ``check`` -- boolean (default: ``True``); Whether to check the input for consistency.

    EXAMPLES::

        sage: WP = WeightedProjectiveSpace(2, ZZ)
        sage: WP(2, 3, 4)
        (2 : 3 : 4)
    """

    def __init__(self, X, v, check: bool = True):
        """
        The Python constructor.

        EXAMPLES::

            sage: WP = WeightedProjectiveSpace([3, 4, 5], QQ)
            sage: P1 = WP(2, 3, 4); P1
            (2 : 3 : 4)
            sage: P2 = WP(2000, 30000, 400000); P2
            (2000 : 30000 : 400000)
            sage: P1 == P2
            True

        The point constructor normalises coordinates at the last position with
        weight `1`::

            sage: WP = WeightedProjectiveSpace([3, 1, 4, 1], QQ)
            sage: P = WP(8, 3, 16, 2); P
            (1 : 3/2 : 1 : 1)
            sage: P == WP(1, 3 / 2, 1, 1)
            True
        """
        super().__init__(X)

        self._normalized = False

        if check:
            # check parent
            from sage.schemes.weighted_projective.weighted_projective_homset import (
                SchemeHomset_points_weighted_projective_ring,
            )

            if not isinstance(X, SchemeHomset_points_weighted_projective_ring):
                raise TypeError(
                    f"ambient space {X} must be a weighted projective space"
                )

            d = X.codomain().ambient_space().ngens()
            if isinstance(v, SchemeMorphism):
                v = list(v)
            else:
                try:
                    if v.parent() in CommutativeRings():
                        v = [v]
                except AttributeError:
                    pass
            if not isinstance(v, (list, tuple)):
                raise TypeError(
                    "argument v (= %s) must be a scheme point, list, or tuple" % str(v)
                )
            if len(v) != d and len(v) != d - 1:
                raise TypeError("v (=%s) must have %s components" % (v, d))

            R = X.value_ring()
            if not R.is_integral_domain():
                raise ValueError(
                    "cannot validate point over a ring that is not an integral domain, "
                    "pass check=False to construct the point"
                )
            v = Sequence(v, R)
            if len(v) == d - 1:  # very common special case
                v.append(R.one())

            # (0 : 0 : ... : 0) is not a valid (weighted) projective point
            if not any(v):
                raise ValueError(
                    f"{v} does not define a valid projective "
                    "point since all entries are zero"
                )

            X.extended_codomain()._check_satisfies_equations(v)

            self._coords = tuple(v)

            # we (try to) normalise coordinates!
            self.normalize_coordinates()

        else:
            self._coords = tuple(v)

    def _repr_(self) -> str:
        return "({})".format(" : ".join(map(repr, self._coords)))

    def _latex_(self) -> str:
        return r"\left({}\right)".format(" : ".join(map(latex, self._coords)))

    def _richcmp_(self, other: SchemeMorphism_point, op) -> bool:
        """
        Test the weighted projective equality of two points.

        INPUT:

        - ``other`` -- a point on weighted projective space

        OUTPUT:

        Boolean

        EXAMPLES::

            sage: WP = WeightedProjectiveSpace([3, 4, 5], QQ)
            sage: u = abs(QQ.random_element()) + 1 / 4 # nonzero
            sage: P, Q = WP(2, 3, 4), WP(2 * u^3, 3 * u^4, 4 * u^5)
            sage: P == Q
            True
            sage: P != Q
            False

        ::

            sage: P, Q = WP(2, 3, 4), WP(2, 3, 5)
            sage: P < Q
            True
            sage: P <= Q
            True
            sage: P > Q
            False
            sage: P >= Q
            False

        In the current implementation, points over different base rings
        are never equal (this may change later)::

            sage: WP(2, 3, 4) == WeightedProjectiveSpace([3, 4, 5], ZZ)(2, 3, 4)
            False
        """
        space = self.codomain()
        if space is not other.codomain():
            return op == op_NE

        if op in (op_EQ, op_NE):
            weights = space.weights()
            # (other[i] / self[i])^(1 / weight[i]) all equal
            # check weights
            if (weights == other.codomain().weights()) != (op == op_EQ):
                return False

            # check zeros
            b1 = all(
                c1 == c2
                for c1, c2 in zip(self._coords, other._coords)
                if c1 == 0 or c2 == 0
            )
            if b1 != (op == op_EQ):
                return False

            # check nonzeros
            prod_weights = prod(weights)
            ratio = [
                (c1 / c2) ** (prod_weights // w)
                for c1, c2, w in zip(self._coords, other._coords, weights)
                if c1 != 0 and c2 != 0
            ]
            r0 = ratio[0]
            b2 = all(r == r0 for r in ratio)
            return b2 == (op == op_EQ)

        return richcmp(self._coords, other._coords, op)

    def __hash__(self):
        """
        Compute the hash value of this point.

        We attempt to normalise the coordinates of this point over the field of
        fractions of the base ring. If this is not possible, return the hash of
        the parent. See :meth:`normalize_coordinates` for more details.

        OUTPUT: Integer.

        .. SEEALSO ::

            :meth:`normalize_coordinates`

        EXAMPLES::

            sage: WP = WeightedProjectiveSpace([1, 3, 1], QQ)
            sage: hash(WP(6, 24, 2)) == hash(WP(3, 3, 1))
            True
            sage: WP = WeightedProjectiveSpace([1, 3, 1], ZZ)
            sage: hash(WP(6, 24, 2)) == hash(WP(3, 3, 1))
            True
        """
        R = self.codomain().base_ring()
        P = self.change_ring(FractionField(R))
        P.normalize_coordinates()
        if P._normalized:
            return hash(tuple(P))
        # if there is no good way to normalize return a constant value
        return hash(self.codomain())

    def scale_by(self, t) -> None:
        """
        Scale the coordinates of the point by ``t``.

        A :exc:`TypeError` occurs if the point is not in the
        base_ring of the codomain after scaling.

        INPUT:

        - ``t`` -- a ring element

        OUTPUT: none

        EXAMPLES::

            sage: WP = WeightedProjectiveSpace([3, 4, 5], ZZ)
            sage: P = WP([8, 16, 32]); P
            (8 : 16 : 32)
            sage: P.scale_by(1 / 2); P
            (1 : 1 : 1)
            sage: P.scale_by(1 / 3); P
            Traceback (most recent call last):
            ...
            TypeError: ...
        """
        if t.is_zero():
            raise ValueError("Cannot scale by 0")
        R = self.codomain().base_ring()
        self._coords = tuple(
            R(u * t**w) for u, w in zip(self._coords, self.codomain().weights())
        )
        self._normalized = False

    def normalize_coordinates(self) -> None:
        """
        Normalise coordinates of this weighted projective point if possible.

        Currently, this method checks if (1) the ambient weighted projective
        space is defined over a field and (2) the weight of any index is `1`.
        If so, the last of which is rescaled to `1`.

        EXAMPLES::

            sage: WP = WeightedProjectiveSpace([3, 1, 5], QQ)
            sage: P = WP([8, 16, 32]); P
            (1/512 : 1 : 1/32768)
            sage: P.scale_by(13); P
            (2197/512 : 13 : 371293/32768)
            sage: P.normalize_coordinates(); P
            (1/512 : 1 : 1/32768)

        ::

            sage: WP = WeightedProjectiveSpace([3, 4, 5], ZZ)
            sage: P = WP([8, 16, 32]); P
            (8 : 16 : 32)
            sage: P.normalize_coordinates(); P
            (8 : 16 : 32)
        """
        if self._normalized:
            return

        if self.base_ring() in Fields():
            weights = self.codomain().weights()
            coords = self._coords
            for i in reversed(range(len(coords))):
                w, c = weights[i], coords[i]
                if w.is_one() and not c.is_zero():
                    # we normalise w.r.t this coordinate
                    self.scale_by(~c)
                    self._normalized = True
                    return



"""
Enumerated sets

Basic enumerated sets
---------------------

- :class:`~sage.combinat.subset.Subsets`, :class:`~sage.combinat.combination.Combinations`
- :class:`~sage.combinat.permutation.Arrangements`, :class:`~sage.combinat.tuple.Tuples`
- :class:`~sage.sets.finite_enumerated_set.FiniteEnumeratedSet`
- :class:`~DisjointUnionEnumeratedSets`

Integer lists
-------------

- :ref:`sage.combinat.partition`
- :ref:`sage.combinat.composition`
- :class:`~sage.combinat.composition_signed.SignedCompositions`
- :class:`IntegerListsLex`
- :ref:`sage.combinat.superpartition`

- :class:`~sage.combinat.integer_vector.IntegerVectors`
- :func:`~sage.combinat.integer_vector_weighted.WeightedIntegerVectors`
- :class:`~sage.combinat.integer_vectors_mod_permgroup.IntegerVectorsModPermutationGroup`

- :ref:`sage.combinat.parking_functions`
- :ref:`sage.combinat.non_decreasing_parking_function`

- :ref:`sage.combinat.sidon_sets`

Words
-----

- :class:`Words`
- :ref:`sage.combinat.subword`
- :ref:`sage.combinat.necklace`
- :ref:`sage.combinat.words.lyndon_word`
- :ref:`sage.combinat.dyck_word`
- :ref:`sage.combinat.debruijn_sequence`
- :ref:`sage.combinat.shuffle`

Permutations and arrangements
-----------------------------

- :ref:`sage.combinat.permutation`
- :ref:`sage.combinat.permutation_cython`
- :ref:`sage.combinat.affine_permutation`
- :class:`~sage.combinat.permutation.Arrangements`
- :ref:`sage.combinat.derangements`
- :ref:`sage.combinat.baxter_permutations`

.. SEEALSO::

    - :class:`SymmetricGroup`, :func:`PermutationGroup`, :ref:`sage.groups.perm_gps.permutation_groups_catalog`
    - :class:`FiniteSetMaps`
    - :ref:`sage.combinat.integer_vectors_mod_permgroup`
    - :ref:`sage.combinat.rsk`

Polyominoes
-----------

- :ref:`sage.combinat.parallelogram_polyomino`

Integer matrices and arrays
---------------------------

- :ref:`sage.combinat.integer_matrices`
- :ref:`sage.combinat.matrices.hadamard_matrix`
- :ref:`sage.combinat.matrices.latin`
- :ref:`sage.combinat.alternating_sign_matrix`
- :ref:`sage.combinat.six_vertex_model`
- :ref:`sage.combinat.similarity_class_type`
- :ref:`sage.combinat.restricted_growth`
- :ref:`sage.combinat.vector_partition`

Subsets and set partitions
--------------------------

- :class:`~sage.combinat.subset.Subsets`, :class:`~sage.combinat.combination.Combinations`
- :class:`~sage.combinat.subsets_pairwise.PairwiseCompatibleSubsets`
- :ref:`sage.combinat.subsets_hereditary`
- :ref:`sage.combinat.set_partition_ordered`
- :ref:`sage.combinat.set_partition`
- :ref:`sage.combinat.diagram_algebras`
- :class:`~sage.combinat.multiset_partition_into_sets_ordered.OrderedMultisetPartitionsIntoSets`,
  :class:`~sage.combinat.multiset_partition_into_sets_ordered.OrderedMultisetPartitionIntoSets`

Trees
-----

- :ref:`sage.combinat.abstract_tree`
- :ref:`sage.combinat.ordered_tree`
- :ref:`sage.combinat.binary_tree`
- :ref:`sage.combinat.rooted_tree`

Enumerated sets related to graphs
---------------------------------

- :ref:`sage.combinat.degree_sequences`
- :ref:`sage.combinat.graph_path`
- :ref:`sage.combinat.perfect_matching`

Miscellaneous enumerated sets
-----------------------------

- :class:`~sage.combinat.integer_lists.invlex.IntegerListsLex`
- :class:`~sage.combinat.integer_vectors_mod_permgroup.IntegerVectorsModPermutationGroup`
- :ref:`sage.combinat.gray_codes`
- :class:`~sage.combinat.gelfand_tsetlin_patterns.GelfandTsetlinPattern`, :class:`~sage.combinat.gelfand_tsetlin_patterns.GelfandTsetlinPatterns`
- :class:`~sage.combinat.knutson_tao_puzzles.KnutsonTaoPuzzleSolver`
- :func:`LatticePolytope`

Backtracking solvers
--------------------

- :func:`~sage.sets.recursively_enumerated_set.RecursivelyEnumeratedSet`
- :class:`~sage.combinat.backtrack.GenericBacktracker`
- :mod:`sage.parallel.map_reduce`
- :ref:`sage.combinat.tiling`
- :ref:`sage.combinat.dlx`
- :ref:`sage.combinat.matrices.dlxcpp`

"""



"""
Gross-Zagier L-series
"""
from sage.lfunctions.pari import lfun_generic, LFunction
from sage.modular.dirichlet import kronecker_character
from sage.modular.modform.l_series_gross_zagier_coeffs import gross_zagier_L_series
from sage.rings.integer import Integer
from sage.structure.sage_object import SageObject


class GrossZagierLseries(SageObject):

    def __init__(self, E, A, prec=53, max_imaginary_part=0) -> None:
        r"""
        Class for the Gross-Zagier `L`-series.

        This is attached to a pair `(E,A)` where `E` is an elliptic curve over
        `\QQ` and `A` is an ideal class in an imaginary quadratic number field.

        For the exact definition, in the more general setting of modular forms
        instead of elliptic curves, see section IV of [GZ1986]_.

        INPUT:

        - ``E`` -- an elliptic curve over `\QQ`

        - ``A`` -- an ideal class in an imaginary quadratic number field

        - ``prec`` -- integer (default: 53); giving the required precision

        - ``max_imaginary_part`` -- real number (default: 0)

        EXAMPLES::

            sage: from sage.modular.modform.l_series_gross_zagier import GrossZagierLseries
            sage: e = EllipticCurve('37a')
            sage: K.<a> = QuadraticField(-40)
            sage: A = K.class_group().gen(0)
            sage: G = GrossZagierLseries(e, A); G
            Gross Zagier L-series attached to Elliptic Curve defined by
            y^2 + y = x^3 - x over Rational Field with ideal class
            Fractional ideal class (2, 1/2*a)

        TESTS::

            sage: K.<b> = QuadraticField(131)
            sage: A = K.class_group().one()
            sage: G = GrossZagierLseries(e, A)
            Traceback (most recent call last):
            ...
            ValueError: A is not an ideal class in an imaginary quadratic field
        """
        self._N = N = E.conductor()
        ideal = A.ideal()
        K = A.gens()[0].parent()
        D = K.disc()
        if not (K.degree() == 2 and D < 0):
            raise ValueError("A is not an ideal class in an"
                             " imaginary quadratic field")
        Q = ideal.quadratic_form().reduced_form()
        epsilon = - kronecker_character(D)(N)

        # first compute the number of required terms
        Lpari = lfun_generic(N**2 * D**2,
                             [0, 0, 1, 1],
                             weight=2, eps=epsilon)
        L = LFunction(Lpari, prec=prec, max_im=max_imaginary_part)
        nterms = Integer(L.cost())
        if nterms > 1e6:
            # just takes way to long
            raise ValueError(f"Too many terms: {nterms}")

        zeta_ord = ideal.number_field().zeta_order()
        an_list = gross_zagier_L_series(E.anlist(nterms + 1), Q, N, zeta_ord)
        Lpari.init_coeffs(an_list[1:])
        self._lfunction = LFunction(Lpari, prec=prec)

        msg = f"Gross Zagier L-series attached to {E} with ideal class {A}"
        self.rename(msg)

    def __call__(self, s, der=0):
        r"""
        Return the value at `s`.

        INPUT:

        - ``s`` -- complex number

        - ``der`` -- order of derivative (default: 0)

        EXAMPLES::

            sage: e = EllipticCurve('37a')
            sage: K.<a> = QuadraticField(-40)
            sage: A = K.class_group().gen(0)
            sage: from sage.modular.modform.l_series_gross_zagier import GrossZagierLseries
            sage: G = GrossZagierLseries(e, A)
            sage: G(3)
            -0.272946890617449
            sage: G(3, 1)
            0.212442670030197
        """
        return self._lfunction.derivative(s, der)

    def taylor_series(self, s=1, series_prec=6, var='z'):
        r"""
        Return the Taylor series at `s`.

        INPUT:

        - ``s`` -- complex number (default: 1)
        - ``series_prec`` -- number of terms (default: 6) in the Taylor series
        - ``var`` -- variable (default: ``'z'``)

        EXAMPLES::

            sage: e = EllipticCurve('37a')
            sage: K.<a> = QuadraticField(-40)
            sage: A = K.class_group().gen(0)
            sage: from sage.modular.modform.l_series_gross_zagier import GrossZagierLseries
            sage: G = GrossZagierLseries(e, A)
            sage: G.taylor_series(2,3)
            -0.613002046122888 + 0.490374999263489*z - 0.122903033710347*z^2 + O(z^3)
        """
        return self._lfunction.taylor_series(s, series_prec, var)



# sage.doctest: needs sage.libs.flint sage.libs.pari
"""
Ambient Hecke modules
"""
# ****************************************************************************
#       Sage: Open Source Mathematical Software
#
#       Copyright (C) 2005 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from . import degenmap
from . import module
from . import submodule

from sage.modules.free_module import FreeModule, FreeModule_generic
from sage.rings.integer import Integer

import sage.arith.misc as arith

from sage.matrix import matrix_space
from sage.matrix.constructor import matrix

from sage.modular.arithgroup.congroup_gamma0 import Gamma0_constructor as Gamma0  # for Sturm bound


class AmbientHeckeModule(module.HeckeModule_free_module):
    """
    An ambient Hecke module, i.e. a Hecke module that is isomorphic as a module
    over its base ring `R` to the standard free module `R^k` for some `k`. This
    is the base class for ambient spaces of modular forms and modular symbols,
    and for Brandt modules.
    """
    def __init__(self, base_ring, rank, level, weight, category=None):
        r"""
        Create an ambient Hecke module.

        EXAMPLES::

            sage: ModularSymbols(6) # indirect doctest
            Modular Symbols space of dimension 3 for Gamma_0(6) of weight 2 with sign 0 over Rational Field
            sage: sage.modular.hecke.ambient_module.AmbientHeckeModule(QQ, 3, 2, 4)
            Generic ambient Hecke module of rank 3, level 2 and weight 4 over Rational Field
        """
        rank = Integer(rank)
        if rank < 0:
            raise ValueError("rank (=%s) must be nonnegative" % rank)
        self.__rank = rank
        module.HeckeModule_free_module.__init__(self, base_ring, level,
                                                weight, category=category)

    def rank(self):
        """
        Return the rank of this ambient Hecke module.

        OUTPUT: integer

        EXAMPLES::

            sage: M = sage.modular.hecke.ambient_module.AmbientHeckeModule(QQ, 3, 11, 2); M
            Generic ambient Hecke module of rank 3, level 11 and weight 2 over Rational Field
            sage: M.rank()
            3
        """
        return self.__rank

    def __add__(self, other):
        r"""
        Return the sum of ``self`` and ``other``.

        As ``self`` is an ambient space, this will only make sense if
        ``other`` is a subspace of ``self``, in which case the answer
        is ``self``.

        EXAMPLES::

            sage: M = ModularSymbols(23)
            sage: M + M is M
            True
            sage: M + 3
            Traceback (most recent call last):
            ...
            TypeError: other (=3) must be a Hecke module
        """
        if not isinstance(other, module.HeckeModule_free_module):
            raise TypeError("other (=%s) must be a Hecke module" % other)
        if other.ambient_hecke_module() == self:
            return self
        raise ArithmeticError("sum only defined for subspaces of a common ambient Hecke module")

    def _repr_(self):
        r"""
        Return the string representation of ``self``.

        This should be overridden by derived classes.

        EXAMPLES::

            sage: sage.modular.hecke.ambient_module.AmbientHeckeModule(QQ, 3, 2, 4)._repr_()
            'Generic ambient Hecke module of rank 3, level 2 and weight 4 over Rational Field'
        """
        return "Generic ambient Hecke module of rank %s, level %s and weight %s over %s" % (self.rank(), self.level(), self.weight(), self.base_ring())

    def _degeneracy_raising_matrix(self, codomain):
        """
        Matrix of the degeneracy map (with t = 1) from ``self`` to ``codomain``,
        whose level should be a multiple of the level of ``self``.

        EXAMPLES::

            sage: sage.modular.hecke.ambient_module.AmbientHeckeModule(QQ, 3, 2, 4)._degeneracy_raising_matrix(4)
            Traceback (most recent call last):
            ...
            NotImplementedError
        """
        raise NotImplementedError

    def _degeneracy_lowering_matrix(self, codomain, t):
        """
        Matrix of the degeneracy map of index t from ``self`` to ``codomain``,
        whose level should be a divisor of the level of ``self``.

        EXAMPLES::

            sage: sage.modular.hecke.ambient_module.AmbientHeckeModule(QQ, 3, 2, 4)._degeneracy_lowering_matrix(2, 2)
            Traceback (most recent call last):
            ...
            NotImplementedError
        """
        raise NotImplementedError

    def _hecke_image_of_ith_basis_element(self, n, i):
        """
        Return the image under the Hecke operator `T_n` of the `i`-th basis
        element.

        EXAMPLES::

            sage: sage.modular.hecke.ambient_module.AmbientHeckeModule(QQ, 3, 2, 4)._hecke_image_of_ith_basis_element(4, 2)
            Traceback (most recent call last):
            ...
            NotImplementedError: All subclasses must implement _compute_hecke_matrix_prime
        """
        return self.hecke_operator(n)(self.gen(i))

    def _set_dual_free_module(self, V):
        r"""
        Store the embedded dual module of this module.

        Since this module is an ambient module, this is not necessary.

        EXAMPLES::

            sage: ModularForms(11, 2)._set_dual_free_module(None)
        """
        pass  # setting dual free module of ambient space is not necessary

    def ambient_hecke_module(self):
        r"""
        Return the ambient space that contains this ambient space.

        This is, of course, just this space again.

        EXAMPLES::

            sage: M = ModularForms(11, 4); M.ambient_hecke_module() is M
            True
        """
        return self

    def complement(self):
        """
        Return the largest Hecke-stable complement of this space.

        EXAMPLES::

            sage: M = ModularSymbols(11,2,1); M
            Modular Symbols space of dimension 2 for Gamma_0(11) of weight 2 with sign 1 over Rational Field
            sage: M.complement()
            Modular Symbols subspace of dimension 0 of Modular Symbols space of dimension 2 for Gamma_0(11) of weight 2 with sign 1 over Rational Field
            sage: C = M.cuspidal_subspace(); C
            Modular Symbols subspace of dimension 1 of Modular Symbols space of dimension 2 for Gamma_0(11) of weight 2 with sign 1 over Rational Field
            sage: C.complement()
            Modular Symbols subspace of dimension 1 of Modular Symbols space of dimension 2 for Gamma_0(11) of weight 2 with sign 1 over Rational Field
        """
        return self.zero_submodule()

    def decomposition_matrix(self):
        r"""
        Return the matrix whose columns form a basis for the canonical
        sorted decomposition of ``self`` coming from the Hecke operators.

        If the simple factors are `D_0, \ldots, D_n`, then the
        first few columns are an echelonized basis for `D_0`, the
        next an echelonized basis for `D_1`, the next for
        `D_2`, etc.

        EXAMPLES::

            sage: S = ModularSymbols(37, 2)
            sage: S.decomposition_matrix()
            [   1    0    0    0 -1/3]
            [   0    1   -1    0  1/2]
            [   0    0    0    1 -1/2]
            [   0    1    1    1    0]
            [   0    0    0    0    1]
        """
        try:
            return self.__decomposition_matrix_cache
        except AttributeError:
            rows = [x.list() for A in self.decomposition()
                    for x in A.basis()]
            A = matrix_space.MatrixSpace(self.base_ring(), self.rank())(rows)
            self.__decomposition_matrix_cache = A
            return self.__decomposition_matrix_cache

    def decomposition_matrix_inverse(self):
        """
        Return the inverse of the decomposition matrix.

        This is the inverse of the matrix returned by
        :meth:`decomposition_matrix`.

        EXAMPLES::

            sage: S = ModularSymbols(37, 2)
            sage: t = S.decomposition_matrix_inverse(); t
            [   1    0    0    0  1/3]
            [   0  1/2 -1/2  1/2 -1/2]
            [   0 -1/2 -1/2  1/2    0]
            [   0    0    1    0  1/2]
            [   0    0    0    0    1]
            sage: t * S.decomposition_matrix() == 1
            True
        """
        try:
            return self.__decomposition_matrix_inverse_cache
        except AttributeError:
            self.__decomposition_matrix_inverse_cache = ~self.decomposition_matrix()
            return self.__decomposition_matrix_inverse_cache

    def degeneracy_map(self, codomain, t=1):
        """
        The `t`-th degeneracy map from ``self`` to the module ``codomain``.

        The level of the codomain must be a divisor or multiple of
        level, and `t` must be a divisor of the quotient.

        INPUT:

        - ``codomain`` -- a Hecke module, which should be of the same type as
          self, or a positive integer (in which case Sage will use
          :meth:`~hecke_module_of_level` to find the "natural" module of the
          corresponding level).
        - ``t`` -- integer; the parameter of the degeneracy map, i.e., the map
          is related to `f(q)` - `f(q^t)`

        OUTPUT: a morphism from ``self`` to ``codomain``

        EXAMPLES::

            sage: M = ModularSymbols(11,sign=1)
            sage: d1 = M.degeneracy_map(33); d1
            Hecke module morphism degeneracy map corresponding to f(q) |--> f(q)
            defined by the matrix
            [ 1  0  0  0 -2 -1]
            [ 0 -1  1  0  0  0]
            Domain:   Modular Symbols space of dimension 2 for Gamma_0(11) of weight ...
            Codomain: Modular Symbols space of dimension 6 for Gamma_0(33) of weight ...
            sage: M.degeneracy_map(33, 3).matrix()
            [ 3  2  0  2 -2  1]
            [ 0  0 -1  1  0  0]
            sage: M = ModularSymbols(33, sign=1)
            sage: d2 = M.degeneracy_map(11); d2.matrix()
            [ 1  0]
            [ 0 -2]
            [ 0  2]
            [ 0  1]
            [-1  0]
            [-1  0]
            sage: (d2*d1).matrix()
            [4 0]
            [0 4]

        ::

            sage: M = ModularSymbols(3, 12, sign=1)
            sage: M.degeneracy_map(1)
            Hecke module morphism degeneracy map corresponding to f(q) |--> f(q)
            defined by the matrix
            [1 0]
            [0 0]
            [0 1]
            [0 1]
            [0 1]
            Domain: Modular Symbols space of dimension 5 for Gamma_0(3) of weight ...
            Codomain: Modular Symbols space of dimension 2 for Gamma_0(1) of weight ...

        ::

            sage: S = M.cuspidal_submodule()
            sage: S.degeneracy_map(1)
            Hecke module morphism defined by the matrix
            [1 0]
            [0 0]
            [0 0]
            Domain: Modular Symbols subspace of dimension 3 of Modular Symbols space ...
            Codomain: Modular Symbols space of dimension 2 for Gamma_0(1) of weight ...

        ::

            sage: D = ModularSymbols(10,4).cuspidal_submodule().decomposition()
            sage: D
            [Modular Symbols subspace of dimension 2 of Modular Symbols space of dimension 10 for Gamma_0(10) of weight 4 with sign 0 over Rational Field,
             Modular Symbols subspace of dimension 4 of Modular Symbols space of dimension 10 for Gamma_0(10) of weight 4 with sign 0 over Rational Field]
            sage: D[1].degeneracy_map(5)
            Hecke module morphism defined by the matrix
            [   0    0   -1    1]
            [   0  1/2  3/2   -2]
            [   0   -1    1    0]
            [   0 -3/4 -1/4    1]
            Domain:   Modular Symbols subspace of dimension 4 of Modular Symbols space ...
            Codomain: Modular Symbols space of dimension 4 for Gamma_0(5) of weight ...

        We check for a subtle caching bug that came up in work on :issue:`10453`::

            sage: loads(dumps(J0(33).decomposition()[0].modular_symbols()))
            Modular Symbols subspace of dimension 2 of
             Modular Symbols space of dimension 9 for
              Gamma_0(33) of weight 2 with sign 0 over Rational Field

        We check that certain absurd inputs are correctly caught::

            sage: chi = kronecker_character(7)
            sage: ModularSymbols(Gamma0(7), 4).degeneracy_map(ModularSymbols(chi, 4))
            Traceback (most recent call last):
            ...
            ValueError: the characters of the domain and codomain must match
        """
        if isinstance(codomain, AmbientHeckeModule):
            M = codomain
            level = int(M.level())
        else:
            level = int(codomain)
            M = None

        t = int(t)

        err = False
        if self.level() % level == 0:
            quo = self.level() // level
            if quo % t != 0:
                err = True
        elif level % self.level() == 0:
            quo = level // self.level()
            if quo % t != 0:
                err = True
        else:
            err = True
        if err:
            raise ValueError(("the level of self (=%s) must be a divisor or multiple of "
                              "level (=%s) and t (=%s) must be a divisor of the quotient") % (self.level(), level, t))

        eps = self.character()
        if eps is not None and level % eps.conductor() != 0:
            raise ArithmeticError("the conductor of the character of this space "
                                  "(=%s) must be divisible by the level (=%s)" % (eps.conductor(), level))

        if M is None:
            M = self.hecke_module_of_level(level)

        if eps is not None and M.character() is not None:
            if eps.primitive_character() != M.character().primitive_character():
                raise ValueError("the characters of the domain and codomain must match")

        key = (M.group(), t)
        # bad idea to use (M, t) as the key, because using complicated objects
        # like modular forms spaces as dictionary keys causes weird behaviour;
        # on the other hand, (M.level(), t) isn't enough information.
        try:
            self._degeneracy_maps
        except AttributeError:
            self._degeneracy_maps = {}

        if key in self._degeneracy_maps:
            return self._degeneracy_maps[key]

        if M.rank() == 0:
            A = matrix_space.MatrixSpace(self.base_ring(), self.rank(), 0)(0)

        elif self.level() % level == 0:  # lower the level
            A = self._degeneracy_lowering_matrix(M, t)

        elif level % self.level() == 0:  # raise the level
            A = self._degeneracy_raising_matrix(M, t)

        d = degenmap.DegeneracyMap(A, self, M, t)
        self._degeneracy_maps[key] = d
        return d

    def dual_free_module(self):
        r"""
        The free module dual to ``self``, as a submodule of the dual
        module of the ambient space. As this space is ambient anyway,
        this just returns ``self.free_module()``.

        EXAMPLES::

            sage: M = ModularForms(2,8); M.dual_free_module()
            Vector space of dimension 3 over Rational Field
            sage: M.dual_free_module() is M.free_module()
            True
        """
        return self.free_module()

    def fcp(self, n, var='x'):
        """
        Return the factorization of the characteristic polynomial of
        the Hecke operator `T_n` of index `n` acting on this space.

        INPUT:

        - ``self`` -- Hecke module invariant under the Hecke operator of index
           `n`

        - ``n`` --- a positive integer.

        - ``var`` --- variable of polynomial (default: ``'x'``)

        OUTPUT:

        - ``list`` -- list of the pairs `(g,e)`, where `g` is an
          irreducible factor of the characteristic polynomial of `T_n`, and `e`
          is its multiplicity.

        EXAMPLES::

            sage: m = ModularSymbols(23, 2, sign=1)
            sage: m.fcp(2)
            (x - 3) * (x^2 + x - 1)
            sage: m.hecke_operator(2).charpoly('x').factor()
            (x - 3) * (x^2 + x - 1)
        """
        n = int(n)
        if n <= 0:
            raise ArithmeticError("n (=%s) must be positive" % n)
        return self.hecke_operator(n).fcp(var)

    def free_module(self):
        """
        Return the free module underlying this ambient Hecke module (the
        forgetful functor from Hecke modules to modules over the base ring)

        EXAMPLES::

            sage: ModularForms(59, 2).free_module()
            Vector space of dimension 6 over Rational Field
        """
        try:
            return self.__free_module
        except AttributeError:
            M = FreeModule(self.base_ring(), self.rank())
            self.__free_module = M
            return M

    def hecke_bound(self):
        r"""
        Return an integer `B` such that the Hecke operators `T_n`, for `n\leq B`,
        generate the full Hecke algebra as a module over the base ring. Note
        that we include the `n` with `n` not coprime to the level.

        At present this returns an unproven guess for non-cuspidal spaces which
        appears to be valid for `M_k(\Gamma_0(N))`, where `k` and `N` are the
        weight and level of ``self``. (It is clearly valid for *cuspidal* spaces
        of any fixed character, as a consequence of the Sturm bound theorem.)
        It returns a hopelessly wrong answer for spaces of full level
        `\Gamma_1`.

        TODO: Get rid of this dreadful bit of code.

        EXAMPLES::

            sage: ModularSymbols(17, 4).hecke_bound()
            15
            sage: ModularSymbols(Gamma1(17), 4).hecke_bound()  # wrong!
            15
        """
        from sage.misc.verbose import verbose
        try:
            if self.is_cuspidal():
                return Gamma0(self.level()).sturm_bound(self.weight())
        except AttributeError:
            pass
        verbose("WARNING: ambient.py -- hecke_bound; returning unproven guess.")
        return Gamma0(self.level()).sturm_bound(self.weight()) + 2 * Gamma0(self.level()).dimension_eis(self.weight()) + 5

    def hecke_module_of_level(self, level):
        r"""
        Return the Hecke module corresponding to ``self`` at the given level, which
        should be either a divisor or a multiple of the level of ``self``.

        This raises :exc:`NotImplementedError`, and should be overridden in
        derived classes.

        EXAMPLES::

            sage: sage.modular.hecke.ambient_module.AmbientHeckeModule.hecke_module_of_level(ModularForms(2, 8),6)
            Traceback (most recent call last):
            ...
            NotImplementedError
        """
        raise NotImplementedError

    def hecke_images(self, i, v):
        """
        Return images of the `i`-th standard basis vector under the
        Hecke operators `T_p` for all integers in `v`.

        INPUT:

        - ``i`` -- nonnegative integer

        - ``v`` -- list of positive integer

        OUTPUT: ``matrix`` -- whose rows are the Hecke images

        EXAMPLES::

            sage: M = ModularSymbols(DirichletGroup(13).0, 3)
            sage: M.T(2)(M.0).element()
            (zeta12 + 4, 0, -1, 1)
            sage: M.hecke_images(0, [1,2])
            [         1          0          0          0]
            [zeta12 + 4          0         -1          1]
        """
        try:
            return self._hecke_images(i, v)
        except (AttributeError, NotImplementedError):
            pass
        # Use slow generic algorithm
        x = self.gen(i)
        X = [self.hecke_operator(n).apply_sparse(x).element() for n in v]
        return matrix(self.base_ring(), X)

    def intersection(self, other):
        """
        Return the intersection of ``self`` and ``other``, which must both lie in
        a common ambient space of modular symbols.

        EXAMPLES::

            sage: M = ModularSymbols(43, sign=1)
            sage: A = M[0] + M[1]
            sage: B = M[1] + M[2]
            sage: A.rank(), B.rank()
            (2, 3)
            sage: C = A.intersection(B); C.rank()  # TODO
            1
        """
        if not isinstance(other, module.HeckeModule_free_module):
            raise TypeError("other (=%s) must be a Hecke module" % other)
        if self.ambient_hecke_module() != other.ambient_hecke_module():
            raise ArithmeticError("intersection only defined for subspaces of a common ambient Hecke module")
        return other  # since self is ambient, so the intersection must equal other.

    def is_ambient(self) -> bool:
        r"""
        Return ``True`` if and only if ``self`` is an ambient Hecke module.

        .. warning::

           ``self`` can only be ambient by being of type
           :class:`AmbientHeckeModule`.

           For example, decomposing a simple ambient space yields a
           single factor, and that factor is *not* considered an
           ambient space.

        EXAMPLES::

            sage: m = ModularSymbols(10)
            sage: m.is_ambient()
            True

        ::

            sage: a = m[0]  # the unique simple factor
            sage: a == m
            True
            sage: a.is_ambient()
            False
        """
        return True

    def is_full_hecke_module(self, compute=True) -> bool:
        """
        Return ``True`` if this space is invariant under the action of
        all Hecke operators, even those that divide the level. This is
        always true for ambient Hecke modules, so return ``True``.

        EXAMPLES::

            sage: ModularSymbols(11, 4).is_full_hecke_module()
            True
        """
        return True

    def is_new(self, p=None) -> bool:
        r"""
        Return ``True`` if this module is entirely new.

        EXAMPLES::

            sage: ModularSymbols(11, 4).is_new()
            False
            sage: ModularSymbols(1, 12).is_new()
            True
        """
        try:
            if p in self.__is_new:
                return self.__is_new[p]
        except AttributeError:
            pass
        AmbientHeckeModule.new_submodule(self, p)
        return self.__is_new[p]

    def is_old(self, p=None) -> bool:
        r"""
        Return ``True`` if this module is entirely old.

        EXAMPLES::

            sage: ModularSymbols(22).is_old()
            True
            sage: ModularSymbols(3, 12).is_old()
            False
        """
        try:
            if p in self.__is_old:
                return self.__is_old[p]
        except AttributeError:
            pass
        self.old_submodule(p)
        return self.__is_old[p]

    def is_submodule(self, V) -> bool:
        """
        Return ``True`` if and only if ``self`` is a submodule of ``V``.

        Since this is an ambient space, this returns ``True`` if and
        only if ``V`` is equal to ``self``.

        EXAMPLES::

            sage: ModularSymbols(1, 4).is_submodule(ModularSymbols(11,4))
            False
            sage: ModularSymbols(11, 4).is_submodule(ModularSymbols(11,4))
            True
        """
        if not isinstance(V, module.HeckeModule_free_module):
            raise TypeError("V must be a Hecke module")
        if not V.is_ambient():
            return False
        return V.ambient_hecke_module() == self

    def linear_combination_of_basis(self, v):
        r"""
        Given a list or vector of length equal to the dimension of ``self``,
        construct the appropriate linear combination of the basis vectors of
        ``self``.

        EXAMPLES::

            sage: ModularForms(3, 12).linear_combination_of_basis([1,0,0,0,1])
            2*q + 2049*q^2 + 177147*q^3 + 4196177*q^4 + 48830556*q^5 + O(q^6)
        """
        return self(v)

    def new_submodule(self, p=None):
        """
        Return the new or `p`-new submodule of ``self``.

        INPUT:

        - ``p`` -- (default: ``None``) if not ``None``, return only
          the `p`-new submodule

        OUTPUT: the new or `p`-new submodule of ``self``, i.e. the intersection of
        the kernel of the degeneracy lowering maps to level `N/p` (for the
        given prime `p`, or for all prime divisors of `N` if `p` is not given).

        If ``self`` is cuspidal, this is a Hecke-invariant complement of the
        corresponding old submodule, but this may break down on Eisenstein
        subspaces (see the amusing example in William Stein's book of a form
        which is new and old at the same time).

        EXAMPLES::

            sage: m = ModularSymbols(33); m.rank()
            9
            sage: m.new_submodule().rank()
            3
            sage: m.new_submodule(3).rank()
            4
            sage: m.new_submodule(11).rank()
            8
        """
        try:
            if self.__is_new[p]:
                return self
        except AttributeError:
            self.__is_new = {}
        except KeyError:
            pass

        if self.rank() == 0:
            self.__is_new[p] = True
            return self
        try:
            return self.__new_submodule[p]
        except AttributeError:
            self.__new_submodule = {}
        except KeyError:
            pass

        # Construct the degeneracy map d.
        N = self.level()
        d = None
        eps = self.character()
        if eps is None:
            f = 1
        else:
            f = eps.conductor()
        if p is None:
            D = arith.prime_divisors(N)
        else:
            if N % p != 0:
                raise ValueError("p must divide the level")
            D = [p]
        for q in D:
            # Here we are only using degeneracy *lowering* maps, so it is fine
            # to be careless and pass an integer for the level. One needs to be
            # a bit more careful with degeneracy *raising* maps for the Gamma1
            # and GammaH cases.
            if ((N // q) % f) == 0:
                NN = N // q
                d1 = self.degeneracy_map(NN, 1).matrix()
                if d is None:
                    d = d1
                else:
                    d = d.augment(d1)
                d = d.augment(self.degeneracy_map(NN, q).matrix())

        if d is None or d == 0:
            self.__is_new[p] = True
            return self
        self.__is_new[p] = False
        ns = self.submodule(d.kernel(), check=False)
        ns.__is_new = {p: True}
        ns._is_full_hecke_module = True
        self.__new_submodule[p] = ns
        return ns

    def nonembedded_free_module(self):
        r"""
        Return the free module corresponding to ``self`` as an abstract free module
        (rather than as a submodule of an ambient free module).

        As this module is ambient anyway, this just returns
        ``self.free_module()``.

        EXAMPLES::

            sage: M = ModularSymbols(11, 2)
            sage: M.nonembedded_free_module() is M.free_module()
            True
        """
        return self.free_module()

    def old_submodule(self, p=None):
        """
        Return the old or `p`-old submodule of ``self``, i.e. the sum of the images
        of the degeneracy maps from level `N/p` (for the given prime `p`, or
        for all primes `p` dividing `N` if `p` is not given).

        INPUT:

        - ``p`` -- (default: ``None``) if not ``None``, return only the `p`-old
          submodule

        OUTPUT: the old or `p`-old submodule of ``self``

        EXAMPLES::

            sage: m = ModularSymbols(33); m.rank()
            9
            sage: m.old_submodule().rank()
            7
            sage: m.old_submodule(3).rank()
            6
            sage: m.new_submodule(11).rank()
            8

        ::

            sage: e = DirichletGroup(16)([-1, 1])
            sage: M = ModularSymbols(e, 3, sign=1); M
            Modular Symbols space of dimension 4 and level 16, weight 3,
             character [-1, 1], sign 1, over Rational Field
            sage: M.old_submodule()
            Modular Symbols subspace of dimension 3 of
             Modular Symbols space of dimension 4 and level 16, weight 3,
              character [-1, 1], sign 1, over Rational Field

        Illustrate that :issue:`10664` is fixed::

            sage: ModularSymbols(DirichletGroup(42)[7], 6, sign=1).old_subspace(3)
            Modular Symbols subspace of dimension 0 of
             Modular Symbols space of dimension 40 and level 42, weight 6,
              character [-1, -1], sign 1, over Rational Field
        """
        try:
            if self.__is_old[p]:
                return self
        except AttributeError:
            self.__is_old = {}
        except KeyError:
            pass

        if self.rank() == 0:
            self.__is_old[p] = True
            return self
        try:
            return self.__old_submodule[p]
        except AttributeError:
            self.__old_submodule = {}
        except KeyError:
            pass

        # Construct the degeneracy map d.
        N = self.level()
        d = None

        eps = self.character()
        if eps is None:
            f = 1
        else:
            f = eps.conductor()

        if p is None:
            D = arith.prime_divisors(N)
        else:
            if N % p != 0:
                raise ValueError("p must divide the level")
            D = [p]

        for q in D:
            NN = N // q
            if NN % f == 0:
                M = self.hecke_module_of_level(NN)

                # Here it is vital to pass self as an argument to
                # degeneracy_map, because M and the level N don't uniquely
                # determine self (e.g. the degeneracy map from level 1 to level
                # N could go to Gamma0(N), Gamma1(N) or anything in between)
                d1 = M.degeneracy_map(self, 1).matrix()

                if d is None:
                    d = d1
                else:
                    d = d.stack(d1)
                d = d.stack(M.degeneracy_map(self, q).matrix())

        if d is None:
            os = self.zero_submodule()
        else:
            os = self.submodule(d.image(), check=False)

        self.__is_old[p] = (os == self)

        os.__is_old = {p: True}
        os._is_full_hecke_module = True
        self.__old_submodule[p] = os
        return os

    def submodule(self, M, Mdual=None, check=True):
        """
        Return the Hecke submodule of ``self`` generated by `M`, which may be a
        submodule of the free module of ``self``, or a list of elements of ``self``.

        EXAMPLES::

            sage: M = ModularForms(37, 2)
            sage: A = M.submodule([M.newforms()[0].element(), M.newforms()[1].element()]); A
            Modular Forms subspace of dimension 2 of
             Modular Forms space of dimension 3 for
              Congruence Subgroup Gamma0(37) of weight 2 over Rational Field
        """
        if check:
            if not isinstance(M, FreeModule_generic):
                V = self.free_module()
                if isinstance(M, (list, tuple)):
                    M = V.span([V(x.element()) for x in M])
                else:
                    M = V.span(M)
            if not M.is_submodule(self.free_module()):
                raise TypeError("M must be a submodule of the free module associated to this module")
            if M == self.free_module():
                return self
        return self._submodule_class()(self, M, Mdual, check=check)

    def _submodule_class(self):
        r"""
        The class of submodules of this module. This is a separate method so it
        can be overridden in derived classes.

        EXAMPLES::

            sage: sage.modular.hecke.ambient_module.AmbientHeckeModule._submodule_class(ModularForms(1, 24))
            <class 'sage.modular.hecke.submodule.HeckeSubmodule'>
            sage: ModularForms(1, 24)._submodule_class()
            <class 'sage.modular.modform.submodule.ModularFormsSubmodule'>
        """
        return submodule.HeckeSubmodule

    def submodule_from_nonembedded_module(self, V, Vdual=None, check=True):
        """
        Create a submodule of this module, from a submodule of an ambient free
        module of the same rank as the rank of ``self``.

        INPUT:

        - ``V`` -- submodule of ambient free module of the same rank as the
          rank of ``self``

        - ``Vdual`` -- used to pass in dual submodule (may be ``None``)

        - ``check`` -- whether to check that submodule is Hecke-equivariant

        OUTPUT: Hecke submodule of ``self``

        EXAMPLES::

            sage: V = QQ^8
            sage: ModularForms(24, 2).submodule_from_nonembedded_module(V.submodule([0]))
            Modular Forms subspace of dimension 0 of
             Modular Forms space of dimension 8 for
              Congruence Subgroup Gamma0(24) of weight 2 over Rational Field
        """
        return self.submodule(V, Vdual, check=check)

    def submodule_generated_by_images(self, M):
        """
        Return the submodule of this ambient modular symbols space
        generated by the images under all degeneracy maps of `M`.

        The space `M` must have the same weight, sign, and group or
        character as this ambient space.

        EXAMPLES::

            sage: ModularSymbols(6, 12).submodule_generated_by_images(ModularSymbols(1,12))
            Modular Symbols subspace of dimension 12 of
             Modular Symbols space of dimension 22 for
              Gamma_0(6) of weight 12 with sign 0 over Rational Field
        """
        S = self.zero_submodule()
        if self.level() % M.level() == 0:
            D = arith.divisors(self.level() // M.level())
        elif M.level() % self.level() == 0:
            D = arith.divisors(M.level() // self.level())
        else:
            D = []
        for t in D:
            d = M.degeneracy_map(self, t)
            if d.codomain() != self:
                raise ArithmeticError("incompatible spaces of modular symbols")
            S += d.image()

        if self.is_full_hecke_module(compute=False):
            S._is_full_hecke_module = True

        return S
